from __future__ import annotations

import copy
import pickle
from collections import namedtuple

import awkward as ak
import awkward_pandas as akpd
import numpy as np
import pandas as pd
import pint
import pytest

import lgdo
from lgdo import Array, VectorOfVectors, lh5

VovColl = namedtuple("VovColl", ["v2d", "v3d", "v4d"])


@pytest.fixture
def testvov():
    v2d = VectorOfVectors([[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]])
    v3d = VectorOfVectors([[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]]])
    v4d = VectorOfVectors(
        [
            [[[1], [2]], [[3, 4], [5]]],
            [[[2, 6]], [[4, 8, 9, 7], [8, 3]]],
            [[[5, 3], [1]]],
        ]
    )

    return VovColl(v2d, v3d, v4d)


def test_init(testvov):
    for v in testvov:
        assert ak.is_valid(v.view_as("ak"))

    assert len(VectorOfVectors()) == 0
    assert len(VectorOfVectors(dtype="ubyte")) == 0
    assert VectorOfVectors(dtype="ubyte").flattened_data.dtype == "ubyte"

    v = VectorOfVectors(
        cumulative_length=np.array([5, 10, 15], dtype="uint32"), dtype="ubyte"
    )
    assert len(v.flattened_data) == 15
    assert len(v[-1]) == 5
    assert v.flattened_data.dtype == "ubyte"
    assert v.cumulative_length.dtype == "uint32"

    v = VectorOfVectors(
        flattened_data=Array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1]),
        cumulative_length=Array([2, 5, 6, 10, 13]),
    )
    assert v == testvov.v2d

    v = VectorOfVectors(ak.Array([[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]]))
    assert v == testvov.v2d

    # shape guess

    v = VectorOfVectors(shape_guess=(10, 20), dtype="int32", fill_val=2)
    assert v.flattened_data == lgdo.Array(shape=(10 * 20,), fill_val=2, dtype="int32")
    assert v.cumulative_length == lgdo.Array(
        np.arange(20, 10 * 20 + 1, 20, dtype="uint32")
    )
    assert ak.is_valid(v.view_as("ak"))

    v = VectorOfVectors(shape_guess=(5, 0), dtype="int32")
    assert v.cumulative_length == lgdo.Array([0, 0, 0, 0, 0])
    assert ak.is_valid(v.view_as("ak"))

    # multi-dimensional
    v = VectorOfVectors(shape_guess=(5, 3, 2), dtype="int16", fill_val=1)
    assert isinstance(v.flattened_data, VectorOfVectors)
    assert isinstance(v.flattened_data.flattened_data, Array)
    assert v.flattened_data.flattened_data.dtype == "int16"
    assert ak.is_valid(v.view_as("ak"))

    assert v.cumulative_length == lgdo.Array([3, 6, 9, 12, 15])
    assert v.flattened_data.cumulative_length == lgdo.Array(
        [i * 2 for i in range(1, 16)]
    )

    assert v == VectorOfVectors(
        [
            [[1, 1], [1, 1], [1, 1]],
            [[1, 1], [1, 1], [1, 1]],
            [[1, 1], [1, 1], [1, 1]],
            [[1, 1], [1, 1], [1, 1]],
            [[1, 1], [1, 1], [1, 1]],
        ],
        dtype="int16",
    )


def test_eq(testvov):
    assert testvov.v2d == VectorOfVectors(
        [[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]]
    )
    assert testvov.v3d == VectorOfVectors(
        [[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]]]
    )
    assert testvov.v4d == VectorOfVectors(
        [
            [[[1], [2]], [[3, 4], [5]]],
            [[[2, 6]], [[4, 8, 9, 7], [8, 3]]],
            [[[5, 3], [1]]],
        ]
    )

    assert VectorOfVectors(
        [[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]], attrs={"ciao": "bello"}
    ) == VectorOfVectors(
        [[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]], attrs={"ciao": "bello"}
    )


def test_len(testvov):
    assert len(testvov.v2d) == 5
    assert len(testvov.v3d) == 3
    assert len(testvov.v4d) == 3

    assert testvov.v2d.ndim == 2
    assert testvov.v3d.ndim == 3
    assert testvov.v4d.ndim == 4


def test_serialization(testvov):
    assert testvov.v2d.cumulative_length == Array([2, 5, 6, 10, 13])
    assert testvov.v2d.flattened_data == Array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1])

    assert testvov.v3d.cumulative_length == Array([2, 4, 5])
    assert isinstance(testvov.v3d.flattened_data, VectorOfVectors)
    assert testvov.v3d.flattened_data.cumulative_length == Array([2, 5, 6, 10, 13])
    assert testvov.v3d.flattened_data.flattened_data == Array(
        [1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1]
    )

    assert isinstance(testvov.v4d.flattened_data, VectorOfVectors)
    assert isinstance(testvov.v4d.flattened_data.flattened_data, VectorOfVectors)
    assert isinstance(testvov.v4d.flattened_data.flattened_data, VectorOfVectors)


def test_datatype_name(testvov):
    for v in testvov:
        assert v.datatype_name() == "array"


def test_form_datatype(testvov):
    assert testvov.v2d.form_datatype() == "array<1>{array<1>{real}}"
    assert testvov.v3d.form_datatype() == "array<1>{array<1>{array<1>{real}}}"
    assert testvov.v4d.form_datatype() == "array<1>{array<1>{array<1>{array<1>{real}}}}"


def test_getitem(testvov):
    testvov = testvov.v2d

    desired = [
        np.array([1, 2]),
        np.array([3, 4, 5]),
        np.array([2]),
        np.array([4, 8, 9, 7]),
        np.array([5, 3, 1]),
    ]

    for i in range(len(desired)):
        assert np.array_equal(desired[i], testvov[i])

    assert np.array_equal(testvov[-1], desired[-1])
    assert np.array_equal(testvov[-2], desired[-2])

    v = VectorOfVectors([[1, 2]], dtype="uint32")
    assert np.array_equal(v[-1], [1, 2])


def test_resize(testvov):
    vov = testvov.v2d

    vov.resize(3)
    assert ak.is_valid(vov.view_as("ak"))
    assert len(vov.cumulative_length) == 3
    assert len(vov.flattened_data) == vov.cumulative_length[-1]
    assert vov == VectorOfVectors([[1, 2], [3, 4, 5], [2]])

    vov.resize(5)
    assert ak.is_valid(vov.view_as("ak"))
    assert len(vov) == 5
    assert len(vov[3]) == 0
    assert len(vov[4]) == 0
    assert vov == VectorOfVectors([[1, 2], [3, 4, 5], [2], [], []])

    vov = testvov.v3d

    vov.resize(3)
    assert ak.is_valid(vov.view_as("ak"))
    assert len(vov.cumulative_length) == 3
    assert len(vov.flattened_data) == vov.cumulative_length[-1]
    assert vov == VectorOfVectors(
        [[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]]]
    )

    vov.resize(5)
    assert ak.is_valid(vov.view_as("ak"))
    assert len(vov) == 5
    assert vov == VectorOfVectors(
        [[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]], [], []]
    )

    v = VectorOfVectors(dtype="i")
    v.resize(3)
    assert ak.is_valid(v.view_as("ak"))
    assert v == VectorOfVectors([[], [], []], dtype="i")


def test_aoesa(testvov):
    testvov = testvov.v2d

    arr = testvov.to_aoesa()
    desired = np.array(
        [
            [1, 2, np.nan, np.nan],
            [3, 4, 5, np.nan],
            [2, np.nan, np.nan, np.nan],
            [4, 8, 9, 7],
            [5, 3, 1, np.nan],
        ]
    )
    assert isinstance(arr, lgdo.ArrayOfEqualSizedArrays)
    assert np.issubdtype(arr.dtype, np.floating)
    assert np.array_equal(arr.nda, desired, True)

    v = VectorOfVectors(
        flattened_data=lgdo.Array(
            nda=np.array([1, 2, 3, 4, 5, 2, 4, 8, 9, 7, 5, 3, 1], dtype="int16")
        ),
        cumulative_length=lgdo.Array(nda=np.array([2, 5, 6, 10, 13])),
    )
    aoesa = v.to_aoesa()

    assert np.issubdtype(aoesa.dtype, np.floating)

    aoesa = v.to_aoesa(fill_val=-999.9)
    assert np.issubdtype(aoesa.nda.dtype, np.floating)

    aoesa = v.to_aoesa(fill_val=-999)
    assert np.issubdtype(aoesa.nda.dtype, np.integer)

    aoesa = v.to_aoesa(fill_val=-999, preserve_dtype=True)
    assert aoesa.nda.dtype == np.int16


def test_set_vector(testvov):
    testvov = testvov.v2d

    testvov[0] = np.zeros(2)
    assert testvov == VectorOfVectors(
        [
            [0, 0],
            [3, 4, 5],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )
    assert ak.is_valid(testvov.view_as("ak"))

    with pytest.raises(ValueError):
        testvov[0] = np.zeros(3)

    testvov[1] = np.zeros(3)

    assert testvov == VectorOfVectors(
        [
            [0, 0],
            [0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )
    assert ak.is_valid(testvov.view_as("ak"))


def test_append(testvov):
    testvov = testvov.v2d

    testvov.append(np.zeros(3))
    assert np.array_equal(testvov[-1], np.zeros(3))

    v = VectorOfVectors(dtype="int64")
    v.append(np.zeros(3))
    assert v == VectorOfVectors([[0, 0, 0]])
    assert ak.is_valid(v.view_as("ak"))


def test_insert(testvov):
    testvov = testvov.v2d

    testvov.insert(2, np.zeros(3))
    assert testvov == VectorOfVectors(
        [
            [1, 2],
            [3, 4, 5],
            [0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )
    assert ak.is_valid(testvov.view_as("ak"))

    v = VectorOfVectors(shape_guess=(3, 5), dtype="int32", fill_val=0)
    v.insert(2, [1, 2, 3])
    assert np.array_equal(v.cumulative_length, [5, 10, 13, 18])
    assert np.array_equal(v[2], [1, 2, 3])


def test_replace(testvov):
    testvov = testvov.v2d

    v = copy.deepcopy(testvov)
    v.replace(1, np.zeros(3))
    assert v == VectorOfVectors(
        [
            [1, 2],
            [0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )
    assert ak.is_valid(v.view_as("ak"))

    v = copy.deepcopy(testvov)
    v.replace(1, np.zeros(2))
    assert v == VectorOfVectors(
        [
            [1, 2],
            [0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )
    assert ak.is_valid(v.view_as("ak"))

    v = copy.deepcopy(testvov)
    v.replace(1, np.zeros(4))
    assert v == VectorOfVectors(
        [
            [1, 2],
            [0, 0, 0, 0],
            [2],
            [4, 8, 9, 7],
            [5, 3, 1],
        ]
    )
    assert ak.is_valid(v.view_as("ak"))


def test_set_vector_unsafe(testvov):
    testvov = testvov.v2d

    desired = [
        np.array([1, 2], dtype=testvov.dtype),
        np.array([3, 4, 5], dtype=testvov.dtype),
        np.array([2], dtype=testvov.dtype),
        np.array([4, 8, 9, 7], dtype=testvov.dtype),
        np.array([5, 3, 1], dtype=testvov.dtype),
    ]
    desired_aoa = np.zeros(shape=(5, 5), dtype=testvov.dtype)
    desired_lens = np.array([len(arr) for arr in desired])

    # test sequential filling
    second_vov = lgdo.VectorOfVectors(shape_guess=(5, 5), dtype=testvov.dtype)
    for i, arr in enumerate(desired):
        second_vov._set_vector_unsafe(i, arr)
        desired_aoa[i, : len(arr)] = arr
    assert testvov == second_vov

    # test vectorized filling
    third_vov = lgdo.VectorOfVectors(shape_guess=(5, 5), dtype=testvov.dtype)
    third_vov._set_vector_unsafe(0, desired_aoa, desired_lens)
    assert testvov == third_vov


def test_iter(testvov):
    testvov = testvov.v2d

    desired = [[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]]

    for c, v in enumerate(testvov):
        assert np.array_equal(v, desired[c])


def test_view_as(testvov):
    v2d = testvov.v2d

    v2d.attrs["units"] = "s"
    with pytest.raises(ValueError):
        v2d.view_as("ak", with_units=True)

    ak_arr = v2d.view_as("ak", with_units=False)

    assert isinstance(ak_arr, ak.Array)
    assert ak.is_valid(ak_arr)
    assert len(ak_arr) == len(v2d)
    assert ak.all(ak_arr == [[1, 2], [3, 4, 5], [2], [4, 8, 9, 7], [5, 3, 1]])

    np_arr = v2d.view_as("np", with_units=True)
    assert isinstance(np_arr, pint.Quantity)
    assert np_arr.u == "second"
    assert isinstance(np_arr.m, np.ndarray)

    np_arr = v2d.view_as("np", with_units=False)
    assert isinstance(np_arr, np.ndarray)
    assert np.issubdtype(np_arr.dtype, np.floating)

    np_arr = v2d.view_as("np", with_units=False, fill_val=0, preserve_dtype=True)
    assert np.issubdtype(np_arr.dtype, np.integer)

    np_arr = v2d.view_as("pd", with_units=False)
    assert isinstance(np_arr, pd.Series)
    assert isinstance(np_arr.ak, akpd.accessor.AwkwardAccessor)

    v3d = testvov.v3d

    ak_arr = v3d.view_as("ak", with_units=False)

    assert isinstance(ak_arr, ak.Array)
    assert ak.is_valid(ak_arr)
    assert len(ak_arr) == len(v3d)
    assert ak.all(ak_arr == [[[1, 2], [3, 4, 5]], [[2], [4, 8, 9, 7]], [[5, 3, 1]]])


def test_lh5_iterator_view_as(lgnd_test_data):
    it = lh5.LH5Iterator(
        lgnd_test_data.get_path("lh5/l200-p03-r000-phy-20230312T055349Z-tier_psp.lh5"),
        "ch1067205/dsp/energies",
    )

    for obj, _, _ in it:
        assert ak.is_valid(obj.view_as("ak"))


def test_pickle(testvov):
    obj = testvov.v2d
    ex = pickle.loads(pickle.dumps(obj))

    desired = [
        np.array([1, 2]),
        np.array([3, 4, 5]),
        np.array([2]),
        np.array([4, 8, 9, 7]),
        np.array([5, 3, 1]),
    ]

    for i in range(len(desired)):
        assert np.array_equal(desired[i], ex[i])



# ruff: noqa: ARG001

from __future__ import annotations

import awkward as ak
import h5py
import numpy as np
import pytest

import lgdo
from lgdo import lh5, types
from lgdo.lh5 import DEFAULT_HDF5_SETTINGS


def test_init():
    lh5.LH5Store()


def test_gimme_file(lgnd_file):
    store = lh5.LH5Store(keep_open=True)

    f = store.gimme_file(lgnd_file)
    assert isinstance(f, h5py.File)
    assert store.files[lgnd_file] == f

    with pytest.raises(FileNotFoundError):
        store.gimme_file("non-existent-file")


def test_gimme_group(lgnd_file, tmptestdir):
    f = h5py.File(lgnd_file)
    store = lh5.LH5Store()
    g = store.gimme_group("/geds", f)
    assert isinstance(g, h5py.Group)

    f = h5py.File(f"{tmptestdir}/testfile.lh5", mode="w")
    g = store.gimme_group("/geds", f, grp_attrs={"attr1": 1}, overwrite=True)
    assert isinstance(g, h5py.Group)


def test_write_objects(lh5_file):
    pass


def test_read_n_rows(lh5_file):
    store = lh5.LH5Store()
    assert store.read_n_rows("/data/struct_full/aoesa", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/array", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/scalar", lh5_file) is None
    assert store.read_n_rows("/data/struct_full/table", lh5_file) == 4
    assert store.read_n_rows("/data/struct_full/voev", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/vov", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/vov3d", lh5_file) == 5
    assert store.read_n_rows("/data/struct_full/wftable", lh5_file) == 10
    assert store.read_n_rows("/data/struct_full/wftable_enc/values", lh5_file) == 10


def test_read_size_in_bytes(lh5_file):
    store = lh5.LH5Store()
    assert store.read_size_in_bytes("/data/struct_full/aoesa", lh5_file) == 100
    assert store.read_size_in_bytes("/data/struct_full/array", lh5_file) == 40
    assert store.read_size_in_bytes("/data/struct_full/scalar", lh5_file) == 8
    assert store.read_size_in_bytes("/data/struct_full/table", lh5_file) == 128
    # assert store.read_size_in_bytes("/data/struct_full/voev", lh5_file) == ?
    assert store.read_size_in_bytes("/data/struct_full/vov", lh5_file) == 144
    assert store.read_size_in_bytes("/data/struct_full/vov3d", lh5_file) == 232
    assert store.read_size_in_bytes("/data/struct_full/wftable", lh5_file) == 20160
    assert (
        store.read_size_in_bytes("/data/struct_full/wftable_enc", lh5_file)
        == 80 + 80 + 40000
    )


def test_get_buffer(lh5_file):
    store = lh5.LH5Store()
    buf = store.get_buffer("/data/struct_full/wftable_enc", lh5_file)
    assert isinstance(buf.values, types.ArrayOfEqualSizedArrays)


def test_read_scalar(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/scalar", lh5_file)
    assert isinstance(lh5_obj, lgdo.Scalar)
    assert lh5_obj.value == 10
    assert n_rows == 1
    assert lh5_obj.attrs["sth"] == 1
    with h5py.File(lh5_file) as h5f:
        assert h5f["/data/struct/scalar"].compression is None


def test_read_array(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/array", lh5_file)
    assert isinstance(lh5_obj, types.Array)
    assert (lh5_obj.nda == np.array([2, 3, 4])).all()
    assert n_rows == 3
    with h5py.File(lh5_file) as h5f:
        assert (
            h5f["/data/struct/array"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )

    lh5_obj, n_rows = store.read("/data/struct_full/array2d", lh5_file)
    assert isinstance(lh5_obj, types.Array)
    assert lh5_obj == types.Array(shape=(23, 56), fill_val=69, dtype=int)


def test_read_array_slice(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read(
        "/data/struct_full/array", lh5_file, start_row=1, n_rows=3
    )
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 3
    assert lh5_obj == lgdo.Array([2, 3, 4])

    lh5_obj, n_rows = store.read(
        "/data/struct_full/array", [lh5_file, lh5_file], start_row=1, n_rows=6
    )
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 6
    assert lh5_obj == lgdo.Array([2, 3, 4, 5, 1, 2])


def test_read_array_fancy_idx(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct_full/array", lh5_file, idx=[0, 3, 4])
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 3
    assert lh5_obj == lgdo.Array([1, 4, 5])

    lh5_obj, n_rows = store.read(
        "/data/struct_full/array", [lh5_file, lh5_file], idx=[[0, 3, 4], [0, 3, 4]]
    )
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 6
    assert lh5_obj == lgdo.Array([1, 4, 5, 1, 4, 5])

    # Test with out of range index
    lh5_obj, n_rows = store.read(
        "/data/struct_full/array", lh5_file, idx=[0, 3, 4, 100]
    )
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 3
    assert lh5_obj == lgdo.Array([1, 4, 5])

    # Test with boolean mask
    lh5_obj, n_rows = store.read(
        "/data/struct_full/array", lh5_file, idx=np.array([1, 0, 0, 1, 1], "bool")
    )
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 3
    assert lh5_obj == lgdo.Array([1, 4, 5])


def test_read_vov(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/vov", lh5_file)
    assert isinstance(lh5_obj, types.VectorOfVectors)

    assert lh5_obj == lgdo.VectorOfVectors(
        [[3, 4, 5], [2], [4, 8, 9, 7]], attrs={"myattr": 2}
    )

    assert n_rows == 3
    assert lh5_obj.attrs["myattr"] == 2

    with h5py.File(lh5_file) as h5f:
        assert (
            h5f["/data/struct/vov/cumulative_length"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )
        assert (
            h5f["/data/struct/vov/flattened_data"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )

    lh5_obj, n_rows = store.read("/data/struct/vov3d", lh5_file)
    assert isinstance(lh5_obj, types.VectorOfVectors)

    assert ak.all(
        lh5_obj.view_as("ak") == ak.Array([[[2], [4, 8, 9, 7]], [[5, 3, 1]], [[3], []]])
    )


def test_read_vov_fancy_idx(lh5_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read("/data/struct_full/vov", lh5_file, idx=[0], n_rows=1)
    assert isinstance(lh5_obj, types.VectorOfVectors)

    lh5_obj, n_rows = store.read("/data/struct_full/vov", lh5_file, idx=[0, 2])
    assert isinstance(lh5_obj, types.VectorOfVectors)
    assert lh5_obj == types.VectorOfVectors([[1, 2], [2]], attrs={"myattr": 2})
    assert n_rows == 2

    lh5_obj, n_rows = store.read("/data/struct_full/vov", lh5_file, idx=[0, 2, 10])
    assert isinstance(lh5_obj, types.VectorOfVectors)
    assert lh5_obj == types.VectorOfVectors([[1, 2], [2]], attrs={"myattr": 2})
    assert n_rows == 2

    lh5_obj, n_rows = store.read("/data/struct_full/vov3d", lh5_file, idx=[0, 2])
    assert isinstance(lh5_obj, types.VectorOfVectors)

    print(lh5_obj)
    assert lh5_obj == types.VectorOfVectors([[[1, 2], [3, 4, 5]], [[5, 3, 1]]])
    assert n_rows == 2


def test_read_voev(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/voev", lh5_file, decompress=False)
    assert isinstance(lh5_obj, types.VectorOfEncodedVectors)

    desired = [np.array([3, 4, 5]), np.array([2]), np.array([4, 8, 9, 7])]

    for i in range(len(desired)):
        assert (desired[i] == lh5_obj[i][0]).all()

    assert n_rows == 3

    lh5_obj, n_rows = store.read(
        "/data/struct/voev", [lh5_file, lh5_file], decompress=False
    )
    assert isinstance(lh5_obj, types.VectorOfEncodedVectors)
    assert n_rows == 6

    with h5py.File(lh5_file) as h5f:
        assert h5f["/data/struct/voev/encoded_data/flattened_data"].compression is None
        assert (
            h5f["/data/struct/voev/encoded_data/cumulative_length"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )
        assert (
            h5f["/data/struct/voev/decoded_size"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )


def test_read_voev_fancy_idx(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read(
        "/data/struct_full/voev", lh5_file, idx=[0, 2], decompress=False
    )
    assert isinstance(lh5_obj, types.VectorOfEncodedVectors)

    desired = [np.array([1, 2]), np.array([2])]

    for i in range(len(desired)):
        assert (desired[i] == lh5_obj[i][0]).all()

    assert n_rows == 2


def test_read_aoesa(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/aoesa", lh5_file)
    assert isinstance(lh5_obj, types.ArrayOfEqualSizedArrays)
    assert (lh5_obj.nda == np.full((3, 5), fill_value=42)).all()


def test_read_aoesa_fancy_idx(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/aoesa", lh5_file, idx=[0, 2])
    assert isinstance(lh5_obj, types.ArrayOfEqualSizedArrays)
    assert (lh5_obj.nda == np.full((2, 5), fill_value=42)).all()

    lh5_obj, n_rows = store.read("/data/struct/aoesa", lh5_file, idx=[0, 2, 10])
    assert isinstance(lh5_obj, types.ArrayOfEqualSizedArrays)
    assert (lh5_obj.nda == np.full((2, 5), fill_value=42)).all()


def test_read_table(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/table", lh5_file)
    assert isinstance(lh5_obj, types.Table)
    assert n_rows == 3

    lh5_obj, n_rows = store.read("/data/struct/table", [lh5_file, lh5_file])
    assert n_rows == 6
    assert lh5_obj.attrs["stuff"] == 5
    assert lh5_obj["a"].attrs["attr"] == 9


def test_read_table_fancy_idx(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/table", lh5_file, idx=[0, 2])
    assert isinstance(lh5_obj, types.Table)
    assert n_rows == 2

    lh5_obj, n_rows = store.read("/data/struct/table", lh5_file, idx=[0, 2, 10])
    assert isinstance(lh5_obj, types.Table)
    assert n_rows == 2

    lh5_obj, n_rows = store.read("/data/struct/table", lh5_file, idx=[])
    assert isinstance(lh5_obj, types.Table)
    assert n_rows == 0


def test_read_empty_struct(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/empty_struct", lh5_file)
    assert isinstance(lh5_obj, types.Struct)
    assert list(lh5_obj.keys()) == []


def test_read_hdf5_compressed_data(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/table", lh5_file)

    assert "compression" not in lh5_obj["b"].attrs
    with h5py.File(lh5_file) as h5f:
        assert (
            h5f["/data/struct/table/a"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )
        assert h5f["/data/struct/table/b"].compression == "gzip"
        assert h5f["/data/struct/table/c"].compression == "gzip"
        assert h5f["/data/struct/table/c"].compression_opts == 9
        assert h5f["/data/struct/table/d"].compression is None


def test_read_wftable(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/wftable", lh5_file)
    assert isinstance(lh5_obj, types.WaveformTable)
    assert n_rows == 3

    lh5_obj, n_rows = store.read("/data/struct/wftable", [lh5_file, lh5_file])
    assert n_rows == 6
    assert lh5_obj.values.attrs["custom"] == 8

    with h5py.File(lh5_file) as h5f:
        assert (
            h5f["/data/struct/wftable/values"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )
        assert (
            h5f["/data/struct/wftable/t0"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )
        assert (
            h5f["/data/struct/wftable/dt"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )


def test_read_wftable_encoded(lh5_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/data/struct/wftable_enc", lh5_file, decompress=False)
    assert isinstance(lh5_obj, types.WaveformTable)
    assert isinstance(lh5_obj.values, types.ArrayOfEncodedEqualSizedArrays)
    assert n_rows == 3
    assert lh5_obj.values.attrs["codec"] == "radware_sigcompress"
    assert "codec_shift" in lh5_obj.values.attrs

    lh5_obj, n_rows = store.read("/data/struct/wftable_enc/values", lh5_file)
    assert isinstance(lh5_obj, lgdo.ArrayOfEqualSizedArrays)
    assert n_rows == 3

    lh5_obj, n_rows = store.read("/data/struct/wftable_enc", lh5_file)
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert isinstance(lh5_obj.values, lgdo.ArrayOfEqualSizedArrays)
    assert n_rows == 3

    lh5_obj_chain, n_rows = store.read(
        "/data/struct/wftable_enc", [lh5_file, lh5_file], decompress=False
    )
    assert n_rows == 6
    assert isinstance(lh5_obj_chain.values, lgdo.ArrayOfEncodedEqualSizedArrays)

    lh5_obj_chain, n_rows = store.read(
        "/data/struct/wftable_enc", [lh5_file, lh5_file], decompress=True
    )
    assert isinstance(lh5_obj_chain.values, lgdo.ArrayOfEqualSizedArrays)
    assert np.array_equal(lh5_obj_chain.values[:3], lh5_obj.values)
    assert np.array_equal(lh5_obj_chain.values[3:], lh5_obj.values)
    assert n_rows == 6

    with h5py.File(lh5_file) as h5f:
        assert (
            h5f[
                "/data/struct/wftable_enc/values/encoded_data/flattened_data"
            ].compression
            is None
        )
        assert h5f["/data/struct/wftable_enc/values/decoded_size"].compression is None
        assert (
            h5f["/data/struct/wftable_enc/t0"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )
        assert (
            h5f["/data/struct/wftable_enc/dt"].compression
            is DEFAULT_HDF5_SETTINGS["compression"]
        )


def test_read_with_field_mask(lh5_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read("/data/struct_full", lh5_file, field_mask=["array"])
    assert list(lh5_obj.keys()) == ["array"]

    lh5_obj, n_rows = store.read(
        "/data/struct_full", lh5_file, field_mask=("array", "table")
    )
    assert sorted(lh5_obj.keys()) == ["array", "table"]

    lh5_obj, n_rows = store.read(
        "/data/struct_full", lh5_file, field_mask={"array": True}
    )
    assert list(lh5_obj.keys()) == ["array"]

    lh5_obj, n_rows = store.read(
        "/data/struct_full", lh5_file, field_mask={"vov": False, "voev": False}
    )
    assert sorted(lh5_obj.keys()) == [
        "aoesa",
        "array",
        "array2d",
        "empty_struct",
        "scalar",
        "table",
        "vov3d",
        "wftable",
        "wftable_enc",
    ]


def test_read_lgnd_array(lgnd_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read("/geds/raw/baseline", lgnd_file)
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 100
    assert len(lh5_obj) == 100

    lh5_obj, n_rows = store.read("/geds/raw/waveform/values", lgnd_file)
    assert isinstance(lh5_obj, types.ArrayOfEqualSizedArrays)


def test_read_lgnd_array_fancy_idx(lgnd_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read(
        "/geds/raw/baseline", lgnd_file, idx=[2, 4, 6, 9, 11, 16, 68]
    )
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 7
    assert len(lh5_obj) == 7
    assert (lh5_obj.nda == [13508, 14353, 14525, 14341, 15079, 11675, 13995]).all()


def test_read_lgnd_vov(lgnd_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read("/geds/raw/tracelist", lgnd_file)
    assert isinstance(lh5_obj, types.VectorOfVectors)
    assert n_rows == 100
    assert len(lh5_obj) == 100


def test_read_lgnd_vov_fancy_idx(lgnd_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read(
        "/geds/raw/tracelist", lgnd_file, idx=[2, 4, 6, 9, 11, 16, 68]
    )
    assert isinstance(lh5_obj, types.VectorOfVectors)
    assert n_rows == 7
    assert len(lh5_obj) == 7
    assert (lh5_obj.cumulative_length.nda == [1, 2, 3, 4, 5, 6, 7]).all()
    assert (lh5_obj.flattened_data.nda == [40, 60, 64, 60, 64, 28, 60]).all()

    lh5_obj, n_rows = store.read("/geds/raw/tracelist", lgnd_file, idx=[])
    assert isinstance(lh5_obj, types.VectorOfVectors)
    assert n_rows == 0
    assert len(lh5_obj) == 0

    lh5_obj, n_rows = store.read("/geds/raw/tracelist", [lgnd_file] * 3, idx=[250])
    assert isinstance(lh5_obj, types.VectorOfVectors)
    assert n_rows == 1
    assert len(lh5_obj) == 1


def test_read_array_concatenation(lgnd_file):
    store = lh5.LH5Store()
    lh5_obj, n_rows = store.read("/geds/raw/baseline", [lgnd_file, lgnd_file])
    assert isinstance(lh5_obj, types.Array)
    assert n_rows == 200
    assert len(lh5_obj) == 200


def test_read_lgnd_waveform_table(lgnd_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read("/geds/raw/waveform", lgnd_file)
    assert isinstance(lh5_obj, types.WaveformTable)

    lh5_obj, n_rows = store.read(
        "/geds/raw/waveform",
        lgnd_file,
        start_row=10,
        n_rows=10,
        field_mask=["t0", "dt"],
    )

    assert isinstance(lh5_obj, types.Table)
    assert list(lh5_obj.keys()) == ["t0", "dt"]
    assert len(lh5_obj) == 10


def test_read_lgnd_waveform_table_fancy_idx(lgnd_file):
    store = lh5.LH5Store()

    lh5_obj, n_rows = store.read(
        "/geds/raw/waveform",
        lgnd_file,
        idx=[7, 9, 25, 27, 33, 38, 46, 52, 57, 59, 67, 71, 72, 82, 90, 92, 93, 94, 97],
    )
    assert isinstance(lh5_obj, types.WaveformTable)
    assert len(lh5_obj) == 19

    lh5_obj, n_rows = store.read(
        "/geds/raw/waveform",
        lgnd_file,
        idx=[
            7,
            9,
            25,
            27,
            33,
            38,
            46,
            52,
            57,
            59,
            67,
            71,
            72,
            82,
            90,
            92,
            93,
            94,
            97,
            1000,
        ],
    )
    assert isinstance(lh5_obj, types.WaveformTable)
    assert len(lh5_obj) == 19


def test_read_compressed_lgnd_waveform_table(lgnd_file, enc_lgnd_file):
    store = lh5.LH5Store()
    wft, _ = store.read("/geds/raw/waveform", enc_lgnd_file)
    assert isinstance(wft.values, types.ArrayOfEqualSizedArrays)
    assert "compression" not in wft.values.attrs



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

import awkward as ak
import numpy as np
import pandas as pd


class LGDO(ABC):
    """Abstract base class representing a LEGEND Data Object (LGDO)."""

    def __new__(cls, *_args, **_kwargs):
        # allow for (un-)pickling LGDO objects.
        obj = super().__new__(cls)
        obj.attrs = {}
        return obj

    @abstractmethod
    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        self.attrs = {} if attrs is None else dict(attrs)

        if "datatype" in self.attrs:
            if self.attrs["datatype"] != self.form_datatype():
                msg = (
                    f"datatype attribute ({self.attrs['datatype']}) does "
                    f"not match class datatype ({self.form_datatype()})!"
                )
                raise ValueError(msg)
        else:
            self.attrs["datatype"] = self.form_datatype()

    @abstractmethod
    def datatype_name(self) -> str:
        """The name for this LGDO's datatype attribute."""

    @abstractmethod
    def form_datatype(self) -> str:
        """Return this LGDO's datatype attribute string."""

    @abstractmethod
    def view_as(
        self, library: str, with_units: bool = False
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        r"""View the LGDO data object as a third-party format data structure.

        This is typically a zero-copy or nearly zero-copy operation unless
        explicitly stated in the concrete LGDO documentation. The view can be
        turned into a copy explicitly by the user with the appropriate methods.
        If requested by the user, the output format supports it and the LGDO
        carries a ``units`` attribute, physical units are attached to the view
        through the :mod:`pint` package.

        Typical supported third-party libraries are:

        - ``pd``: :mod:`pandas`
        - ``np``: :mod:`numpy`
        - ``ak``: :mod:`awkward`

        Note
        ----
        Awkward does not support attaching units through Pint, at the moment.

        but the actual supported formats may vary depending on the concrete
        LGDO class.

        Parameters
        ----------
        library
            format of the returned data view.
        with_units
            forward physical units to the output data.
        """

    def getattrs(self, datatype: bool = False) -> dict:
        """Return a copy of the LGDO attributes dictionary.

        Parameters
        ----------
        datatype
            if ``False``, remove ``datatype`` attribute from the output
            dictionary.
        """
        d = dict(self.attrs)
        if not datatype:
            d.pop("datatype", None)
        return d

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"(attrs={self.attrs!r})"



from __future__ import annotations

import pytest

from lgdo import lh5


@pytest.fixture
def wftable(lgnd_test_data):
    store = lh5.LH5Store()
    wft, _ = store.read(
        "/geds/raw/waveform",
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
    )
    return wft



from __future__ import annotations

import pickle

import pytest

import lgdo


def test_datatype_name():
    struct = lgdo.Struct()
    assert struct.datatype_name() == "struct"


def test_form_datatype():
    struct = lgdo.Struct()
    assert struct.form_datatype() == "struct{}"


def test_update_datatype():
    struct = lgdo.Struct()
    struct.update_datatype()
    assert struct.attrs["datatype"] == "struct{}"


def test_init():
    obj_dict = {"scalar1": lgdo.Scalar(value=10)}
    attrs = {"attr1": 1}
    struct = lgdo.Struct(obj_dict=obj_dict, attrs=attrs)
    assert dict(struct) == obj_dict
    assert struct.attrs == attrs | {"datatype": "struct{scalar1}"}

    with pytest.raises(ValueError):
        lgdo.Struct(obj_dict={"scalar1": 1})

    with pytest.raises(ValueError):
        lgdo.Struct(obj_dict={"scalar1": lgdo.Scalar(value=10), "thing": int})


def test_init_nested():
    obj_dict = {
        "scalar1": lgdo.Scalar(10),
        "struct1": {"field1": lgdo.Scalar(11), "field2": lgdo.Array([1, 2, 3, 4])},
    }
    struct = lgdo.Struct(obj_dict)
    assert isinstance(struct.struct1, lgdo.Struct)
    assert isinstance(struct.struct1.field1, lgdo.Scalar)
    assert struct.struct1.field1.value == 11
    assert isinstance(struct.struct1.field2, lgdo.Array)
    assert struct.struct1.field2 == lgdo.Array([1, 2, 3, 4])


def test_add_field():
    struct = lgdo.Struct()
    struct.add_field("scalar1", lgdo.Scalar(value=10))

    assert struct.attrs["datatype"] == "struct{scalar1}"
    assert struct["scalar1"].__class__.__name__ == "Scalar"

    struct.add_field("array1", lgdo.Array(shape=(700, 21), dtype="f", fill_val=2))
    assert struct.attrs["datatype"] == "struct{array1,scalar1}"

    struct["array2"] = lgdo.Array(shape=(700, 21), dtype="f", fill_val=2)
    assert struct.attrs["datatype"] == "struct{array1,array2,scalar1}"


def test_getattr():
    struct = lgdo.Struct()
    struct["scalar1"] = lgdo.Scalar(value=10)
    assert struct.scalar1.value == 10


def test_remove_field():
    struct = lgdo.Struct()
    struct.add_field("scalar1", lgdo.Scalar(value=10))
    struct.add_field("array1", lgdo.Array(shape=(10), fill_val=0))

    struct.remove_field("scalar1")
    assert list(struct.keys()) == ["array1"]

    struct.remove_field("array1", delete=True)
    assert list(struct.keys()) == []


def test_pickle():
    obj_dict = {"scalar1": lgdo.Scalar(value=10)}
    attrs = {"attr1": 1}
    struct = lgdo.Struct(obj_dict=obj_dict, attrs=attrs)

    ex = pickle.loads(pickle.dumps(struct))
    assert isinstance(ex, lgdo.Struct)
    assert ex.attrs["attr1"] == 1
    assert ex.attrs["datatype"] == struct.attrs["datatype"]
    assert ex["scalar1"].value == 10



""":class:`~.lgdo.typing.vectorofvectors.VectorOfVectors` utilities."""

from __future__ import annotations

import logging
from collections.abc import Sequence

import awkward as ak
import numba
import numpy as np
from numpy.typing import NDArray

from ..utils import numba_defaults_kwargs as nb_kwargs
from .array import Array

log = logging.getLogger(__name__)


def build_cl(
    sorted_array_in: NDArray, cumulative_length_out: NDArray | None = None
) -> NDArray:
    """Build a cumulative length array from an array of sorted data.

    Examples
    --------
    >>> build_cl(np.array([3, 3, 3, 4])
    array([3., 4.])

    For a `sorted_array_in` of indices, this is the inverse of
    :func:`.explode_cl`, in the sense that doing
    ``build_cl(explode_cl(cumulative_length))`` would recover the original
    `cumulative_length`.

    Parameters
    ----------
    sorted_array_in
        array of data already sorted; each N matching contiguous entries will
        be converted into a new row of `cumulative_length_out`.
    cumulative_length_out
        a pre-allocated array for the output `cumulative_length`. It will
        always have length <= `sorted_array_in`, so giving them the same length
        is safe if there is not a better guess.

    Returns
    -------
    cumulative_length_out
        the output cumulative length array. If the user provides a
        `cumulative_length_out` that is too long, this return value is sliced
        to contain only the used portion of the allocated memory.
    """
    if len(sorted_array_in) == 0:
        return None
    sorted_array_in = np.asarray(sorted_array_in)
    if cumulative_length_out is None:
        cumulative_length_out = np.zeros(len(sorted_array_in), dtype=np.uint64)
    else:
        cumulative_length_out.fill(0)
    if len(cumulative_length_out) == 0 and len(sorted_array_in) > 0:
        msg = "cumulative_length_out too short ({len(cumulative_length_out)})"
        raise ValueError(msg)
    return _nb_build_cl(sorted_array_in, cumulative_length_out)


@numba.njit(**nb_kwargs)
def _nb_build_cl(sorted_array_in: NDArray, cumulative_length_out: NDArray) -> NDArray:
    """numbified inner loop for build_cl"""
    ii = 0
    last_val = sorted_array_in[0]
    for val in sorted_array_in:
        if val != last_val:
            ii += 1
            cumulative_length_out[ii] = cumulative_length_out[ii - 1]
            if ii >= len(cumulative_length_out):
                msg = "cumulative_length_out too short"
                raise RuntimeError(msg)
            last_val = val
        cumulative_length_out[ii] += 1
    ii += 1
    return cumulative_length_out[:ii]


@numba.guvectorize(
    [
        f"{data_type}[:,:],{size_type}[:],{data_type}[:]"
        for data_type in [
            "b1",
            "i1",
            "i2",
            "i4",
            "i8",
            "u1",
            "u2",
            "u4",
            "u8",
            "f4",
            "f8",
            "c8",
            "c16",
        ]
        for size_type in ["i4", "i8", "u4", "u8"]
    ],
    "(l,m),(l),(n)",
    **nb_kwargs,
)
def _nb_fill(aoa_in: NDArray, len_in: NDArray, flattened_array_out: NDArray):
    """Vectorized function to fill flattened array from array of arrays and
    lengths. Values in aoa_in past lengths will not be copied.

    Parameters
    ----------
    aoa_in
        array of arrays containing values to be copied
    len_in
        array of vector lengths for each row of aoa_in
    flattened_array_out
        flattened array to copy values into. Must be longer than sum of
        lengths in len_in
    """

    if len(flattened_array_out) < len_in.sum():
        msg = "flattened array not large enough to hold values"
        raise ValueError(msg)

    start = 0
    for i, ll in enumerate(len_in):
        stop = start + ll
        flattened_array_out[start:stop] = aoa_in[i, :ll]
        start = stop


def explode_cl(cumulative_length: NDArray, array_out: NDArray | None = None) -> NDArray:
    """Explode a `cumulative_length` array.

    Examples
    --------
    >>> explode_cl(np.array([2, 3]))
    array([0., 0., 1.])

    This is the inverse of :func:`.build_cl`, in the sense that doing
    ``build_cl(explode_cl(cumulative_length))`` would recover the original
    `cumulative_length`.

    Parameters
    ----------
    cumulative_length
        the cumulative length array to be exploded.
    array_out
        a pre-allocated array to hold the exploded cumulative length array.
        The length should be equal to ``cumulative_length[-1]``.

    Returns
    -------
    array_out
        the exploded cumulative length array.
    """
    cumulative_length = np.asarray(cumulative_length)
    out_len = cumulative_length[-1] if len(cumulative_length) > 0 else 0
    if array_out is None:
        array_out = np.empty(int(out_len), dtype=np.uint64)
    if len(array_out) != out_len:
        msg = f"bad lengths: cl[-1] ({cumulative_length[-1]}) != out ({len(array_out)})"
        raise ValueError(msg)
    return _nb_explode_cl(cumulative_length, array_out)


@numba.njit(**nb_kwargs)
def _nb_explode_cl(cumulative_length: NDArray, array_out: NDArray) -> NDArray:
    """numbified inner loop for explode_cl"""
    out_len = cumulative_length[-1] if len(cumulative_length) > 0 else 0
    if len(array_out) != out_len:
        msg = "bad lengths"
        raise ValueError(msg)
    start = 0
    for ii in range(len(cumulative_length)):
        nn = int(cumulative_length[ii] - start)
        for jj in range(nn):
            array_out[int(start + jj)] = ii
        start = cumulative_length[ii]
    return array_out


def explode(
    cumulative_length: NDArray, array_in: NDArray, array_out: NDArray | None = None
) -> NDArray:
    """Explode a data array using a `cumulative_length` array.

    This is identical to :func:`.explode_cl`, except `array_in` gets exploded
    instead of `cumulative_length`.

    Examples
    --------
    >>> explode(np.array([2, 3]), np.array([3, 4]))
    array([3., 3., 4.])

    Parameters
    ----------
    cumulative_length
        the cumulative length array to use for exploding.
    array_in
        the data to be exploded. Must have same length as `cumulative_length`.
    array_out
        a pre-allocated array to hold the exploded data. The length should be
        equal to ``cumulative_length[-1]``.

    Returns
    -------
    array_out
        the exploded cumulative length array.
    """
    cumulative_length = np.asarray(cumulative_length)
    array_in = np.asarray(array_in)
    out_len = cumulative_length[-1] if len(cumulative_length) > 0 else 0
    if array_out is None:
        array_out = np.empty(out_len, dtype=array_in.dtype)
    if len(cumulative_length) != len(array_in) or len(array_out) != out_len:
        msg = (
            f"bad lengths: cl ({len(cumulative_length)}) != in ({len(array_in)}) "
            f"and cl[-1] ({cumulative_length[-1]}) != out ({len(array_out)})"
        )
        raise ValueError(msg)
    return _nb_explode(cumulative_length, array_in, array_out)


@numba.njit(**nb_kwargs)
def _nb_explode(
    cumulative_length: NDArray, array_in: NDArray, array_out: NDArray
) -> NDArray:
    """Numbified inner loop for :func:`.explode`."""
    out_len = cumulative_length[-1] if len(cumulative_length) > 0 else 0
    if len(cumulative_length) != len(array_in) or len(array_out) != out_len:
        msg = "bad lengths"
        raise ValueError(msg)
    ii = 0
    for jj in range(len(array_out)):
        while ii < len(cumulative_length) and jj >= cumulative_length[ii]:
            ii += 1
        array_out[jj] = array_in[ii]
    return array_out


def explode_arrays(
    cumulative_length: Array,
    arrays: Sequence[NDArray],
    arrays_out: Sequence[NDArray] | None = None,
) -> list:
    """Explode a set of arrays using a `cumulative_length` array.

    Parameters
    ----------
    cumulative_length
        the cumulative length array to use for exploding.
    arrays
        the data arrays to be exploded. Each array must have same length as
        `cumulative_length`.
    arrays_out
        a list of pre-allocated arrays to hold the exploded data. The length of
        the list should be equal to the length of `arrays`, and each entry in
        arrays_out should have length ``cumulative_length[-1]``. If not
        provided, output arrays are allocated for the user.

    Returns
    -------
    arrays_out
        the list of exploded cumulative length arrays.
    """
    cumulative_length = np.asarray(cumulative_length)
    for ii in range(len(arrays)):
        arrays[ii] = np.asarray(arrays[ii])
    out_len = cumulative_length[-1] if len(cumulative_length) > 0 else 0
    if arrays_out is None:
        arrays_out = []
        for array in arrays:
            arrays_out.append(np.empty(out_len, dtype=array.dtype))
    for ii in range(len(arrays)):
        explode(cumulative_length, arrays[ii], arrays_out[ii])
    return arrays_out


def _ak_is_jagged(type_: ak.types.Type) -> bool:
    """Returns ``True`` if :class:`ak.Array` is jagged at all axes.

    This assures that :func:`ak.to_buffers` returns the expected data
    structures.
    """
    if isinstance(type_, ak.Array):
        return _ak_is_jagged(type_.type)

    if isinstance(type_, (ak.types.ArrayType, ak.types.ListType)):
        return _ak_is_jagged(type_.content)

    if isinstance(type_, ak.types.ScalarType):
        msg = "Expected ArrayType or its content"
        raise TypeError(msg)

    return not isinstance(type_, ak.types.RegularType)


# https://github.com/scikit-hep/awkward/discussions/3049
def _ak_is_valid(type_: ak.types.Type) -> bool:
    """Returns ``True`` if :class:`ak.Array` contains only elements we can serialize to LH5."""
    if isinstance(type_, ak.Array):
        return _ak_is_valid(type_.type)

    if isinstance(type_, (ak.types.ArrayType, ak.types.ListType)):
        return _ak_is_valid(type_.content)

    if isinstance(type_, ak.types.ScalarType):
        msg = "Expected ArrayType or its content"
        raise TypeError(msg)

    return not isinstance(
        type_,
        (
            ak.types.OptionType,
            ak.types.UnionType,
            ak.types.RecordType,
        ),
    )

    return isinstance(type_, ak.types.NumpyType)



from __future__ import annotations

import os

import pytest

from lgdo.lh5 import utils


@pytest.fixture(scope="module")
def lgnd_file(lgnd_test_data):
    return lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5")


def test_expand_vars():
    # Check env variable expansion
    os.environ["PYGAMATESTBASEDIR"] = "a_random_string"
    assert utils.expand_vars("$PYGAMATESTBASEDIR/blah") == "a_random_string/blah"

    # Check user variable expansion
    assert (
        utils.expand_vars(
            "$PYGAMATESTBASEDIR2/blah",
            substitute={"PYGAMATESTBASEDIR2": "a_random_string"},
        )
        == "a_random_string/blah"
    )


def test_expand_path(lgnd_test_data):
    files = [
        lgnd_test_data.get_path(
            "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_dsp.lh5"
        ),
        lgnd_test_data.get_path(
            "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r001/l200-p03-r001-cal-20230318T012228Z-tier_dsp.lh5"
        ),
    ]
    base_dir = os.path.dirname(files[0])

    assert utils.expand_path(f"{base_dir}/*20230318T012144Z*") == files[0]

    # Should fail if file not found
    with pytest.raises(FileNotFoundError):
        utils.expand_path(f"{base_dir}/not_a_real_file.lh5")

    # Should fail if multiple files found
    with pytest.raises(FileNotFoundError):
        utils.expand_path(f"{base_dir}/*.lh5")

    # Check if it finds a list of files correctly
    assert sorted(utils.expand_path(f"{base_dir}/*.lh5", list=True)) == sorted(files)



from __future__ import annotations

import logging

import numpy as np

from .... import types
from ... import utils
from ...exceptions import LH5EncodeError
from .array import _h5_write_array

log = logging.getLogger(__name__)


def _h5_write_vector_of_vectors(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    assert isinstance(obj, types.VectorOfVectors)

    group = utils.get_h5_group(
        name, group, grp_attrs=obj.attrs, overwrite=(wo_mode == "o")
    )
    if n_rows is None or n_rows > obj.cumulative_length.nda.shape[0] - start_row:
        n_rows = obj.cumulative_length.nda.shape[0] - start_row

    # if appending we need to add an appropriate offset to the
    # cumulative lengths as appropriate for the in-file object
    # declare here because we have to subtract it off at the end
    offset = np.int64(0)
    if (wo_mode in ("a", "o")) and "cumulative_length" in group:
        len_cl = len(group["cumulative_length"])
        # if append, ignore write_start and set it to total number of vectors
        if wo_mode == "a":
            write_start = len_cl
        if len_cl > 0:
            # set offset to correct number of elements in flattened_data until write_start
            offset = group["cumulative_length"][write_start - 1]

    # First write flattened_data array. Only write rows with data.
    fd_start = 0 if start_row == 0 else obj.cumulative_length.nda[start_row - 1]
    fd_n_rows = (
        obj.cumulative_length.nda[start_row + n_rows - 1] - fd_start
        if len(obj.cumulative_length) > 0
        else 0
    )

    if isinstance(obj.flattened_data, types.Array):
        _func = _h5_write_array
    elif isinstance(obj.flattened_data, types.VectorOfVectors):
        _func = _h5_write_vector_of_vectors
    else:
        msg = (
            "don't know how to serialize to disk flattened_data "
            "of {type(obj.flattened_data).__name__} type"
        )
        raise LH5EncodeError(msg, lh5_file, group, f"{name}.flattened_data")

    _func(
        obj.flattened_data,
        "flattened_data",
        lh5_file,
        group=group,
        start_row=fd_start,
        n_rows=fd_n_rows,
        wo_mode=wo_mode,
        write_start=offset,
        **h5py_kwargs,
    )

    # now offset is used to give appropriate in-file values for
    # cumulative_length. Need to adjust it for start_row, if different from zero
    if start_row > 0:
        offset -= obj.cumulative_length.nda[start_row - 1]

    # Add offset to obj.cumulative_length itself to avoid memory allocation.
    # Then subtract it off after writing! (otherwise it will be changed
    # upon return)

    # NOTE: this operation is not numerically safe (uint overflow in the lower
    # part of the array), but this is not a problem because those values are
    # not written to disk and we are going to restore the offset at the end
    np.add(
        obj.cumulative_length.nda,
        offset,
        out=obj.cumulative_length.nda,
        casting="unsafe",
    )

    _h5_write_array(
        obj.cumulative_length,
        "cumulative_length",
        lh5_file,
        group=group,
        start_row=start_row,
        n_rows=n_rows,
        wo_mode=wo_mode,
        write_start=write_start,
        **h5py_kwargs,
    )

    np.subtract(
        obj.cumulative_length.nda,
        offset,
        out=obj.cumulative_length.nda,
        casting="unsafe",
    )



# ruff: noqa: ARG001

from __future__ import annotations

import numpy as np
import pytest

import lgdo
from lgdo import lh5


@pytest.fixture(scope="module")
def lgnd_file(lgnd_test_data):
    return lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5")


def test_basics(lgnd_file):
    lh5_it = lh5.LH5Iterator(
        lgnd_file,
        "/geds/raw",
        entry_list=range(100),
        field_mask=["baseline"],
        buffer_len=5,
    )

    lh5_obj, n_rows = lh5_it.read(4)
    assert n_rows == 5
    assert isinstance(lh5_obj, lgdo.Table)
    assert list(lh5_obj.keys()) == ["baseline"]
    assert (
        lh5_obj["baseline"].nda == np.array([14353, 14254, 14525, 11656, 13576])
    ).all()

    for lh5_obj, entry, n_rows in lh5_it:
        assert len(lh5_obj) == 5
        assert n_rows == 5
        assert entry % 5 == 0
        assert all(lh5_it.current_local_entries == np.arange(entry, entry + 5))
        assert all(lh5_it.current_global_entries == np.arange(entry, entry + 5))
        assert all(lh5_it.current_files == [lgnd_file] * 5)
        assert all(lh5_it.current_groups == ["/geds/raw"] * 5)


def test_errors(lgnd_file):
    with pytest.raises(RuntimeError):
        lh5.LH5Iterator("non-existent-file.lh5", "random-group")

    with pytest.raises(ValueError):
        lh5.LH5Iterator(1, 2)

    with pytest.raises(ValueError):
        lh5.LH5Iterator(
            lgnd_file,
            "/geds/raw",
            entry_list=range(100),
            entry_mask=np.ones(100, "bool"),
        )


def test_lgnd_waveform_table_fancy_idx(lgnd_file):
    lh5_it = lh5.LH5Iterator(
        lgnd_file,
        "geds/raw/waveform",
        entry_list=[
            7,
            9,
            25,
            27,
            33,
            38,
            46,
            52,
            57,
            59,
            67,
            71,
            72,
            82,
            90,
            92,
            93,
            94,
            97,
        ],
        buffer_len=5,
    )

    lh5_obj, n_rows = lh5_it.read(0)
    assert isinstance(lh5_obj, lgdo.WaveformTable)
    assert len(lh5_obj) == 5


@pytest.fixture(scope="module")
def more_lgnd_files(lgnd_test_data):
    return [
        [
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
            ),
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012228Z-tier_raw.lh5"
            ),
        ],
        [
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/hit/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_hit.lh5"
            ),
            lgnd_test_data.get_path(
                "lh5/prod-ref-l200/generated/tier/hit/cal/p03/r001/l200-p03-r001-cal-20230318T012228Z-tier_hit.lh5"
            ),
        ],
    ]


def test_friend(more_lgnd_files):
    lh5_raw_it = lh5.LH5Iterator(
        more_lgnd_files[0],
        "ch1084803/raw",
        field_mask=["waveform", "baseline"],
        buffer_len=5,
    )
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        "ch1084803/hit",
        field_mask=["is_valid_0vbb"],
        buffer_len=5,
        friend=lh5_raw_it,
    )

    lh5_obj, n_rows = lh5_it.read(0)

    assert n_rows == 5
    assert isinstance(lh5_obj, lgdo.Table)
    assert set(lh5_obj.keys()) == {"waveform", "baseline", "is_valid_0vbb"}


def test_iterate(more_lgnd_files):
    # iterate through all hit groups in all files; there are 10 entries in
    # each group/file
    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        entry_list=[0, 1, 2, 3, 5, 8, 13, 21, 34, 55],
        buffer_len=5,
    )

    exp_loc_entries = [[0, 1, 2, 3, 5], [8, 3, 1, 4, 5]]
    exp_glob_entries = [[0, 1, 2, 3, 5], [8, 13, 21, 34, 55]]
    exp_files = [
        [more_lgnd_files[1][0]] * 5,
        [more_lgnd_files[1][0]] * 3 + [more_lgnd_files[1][1]] * 2,
    ]
    exp_groups = [
        ["ch1084803/hit"] * 5,
        [
            "ch1084803/hit",
            "ch1084804/hit",
            "ch1121600/hit",
            "ch1084803/hit",
            "ch1121600/hit",
        ],
    ]

    for lh5_out, entry, n_rows in lh5_it:
        assert set(lh5_out.keys()) == {"is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"}
        assert entry % 5 == 0
        assert n_rows == 5
        assert all(lh5_it.current_local_entries == exp_loc_entries[entry // 5])
        assert all(lh5_it.current_global_entries == exp_glob_entries[entry // 5])
        assert all(lh5_it.current_files == exp_files[entry // 5])
        assert all(lh5_it.current_groups == exp_groups[entry // 5])

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        ["ch1084803/hit", "ch1084804/hit", "ch1121600/hit"],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        entry_list=[[0, 1, 2, 3, 5, 8], [3], [1], [4], [], [5]],
        buffer_len=5,
    )

    for lh5_out, entry, n_rows in lh5_it:
        assert set(lh5_out.keys()) == {"is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"}
        assert entry % 5 == 0
        assert n_rows == 5
    print(lh5_it.get_global_entrylist())
    assert all(
        i == j
        for i, j in zip(
            lh5_it.get_global_entrylist(), [0, 1, 2, 3, 5, 8, 13, 21, 34, 55]
        )
    )

    lh5_it = lh5.LH5Iterator(
        more_lgnd_files[1],
        [["ch1084803/hit", "ch1084804/hit"], ["ch1084803/hit", "ch1121600/hit"]],
        field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
        buffer_len=5,
    )

    for lh5_out, entry, n_rows in lh5_it:
        assert set(lh5_out.keys()) == {"is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"}
        assert entry % 5 == 0
        assert n_rows == 5

    with pytest.raises(ValueError):
        lh5.LH5Iterator(
            more_lgnd_files[1],
            [["ch1084803/hit", "ch1084804/hit"], "ch1084803/hit"],
            field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
            buffer_len=5,
        )

    with pytest.raises(ValueError):
        lh5.LH5Iterator(
            more_lgnd_files[1],
            [["ch1084803/hit"], ["ch1084804/hit"], ["ch1084803/hit"]],
            field_mask=["is_valid_0vbb", "timestamp", "zacEmax_ctc_cal"],
            buffer_len=5,
        )



from __future__ import annotations

import logging

from ....types import Array, ArrayOfEqualSizedArrays, FixedSizeArray
from . import utils
from .ndarray import _h5_read_ndarray

log = logging.getLogger(__name__)


def _h5_read_array_generic(type_, h5d, fname, oname, **kwargs):
    nda, attrs, n_rows_to_read = _h5_read_ndarray(h5d, fname, oname, **kwargs)

    obj_buf = kwargs["obj_buf"]

    if obj_buf is None:
        return type_(nda=nda, attrs=attrs), n_rows_to_read

    utils.check_obj_buf_attrs(obj_buf.attrs, attrs, fname, oname)

    return obj_buf, n_rows_to_read


def _h5_read_array(h5d, fname, oname, **kwargs):
    return _h5_read_array_generic(Array, h5d, fname, oname, **kwargs)


def _h5_read_fixedsize_array(h5d, fname, oname, **kwargs):
    return _h5_read_array_generic(FixedSizeArray, h5d, fname, oname, **kwargs)


def _h5_read_array_of_equalsized_arrays(h5d, fname, oname, **kwargs):
    return _h5_read_array_generic(ArrayOfEqualSizedArrays, h5d, fname, oname, **kwargs)



"""
.. warning::
    This subpackage is deprecated, use :mod:`lgdo.lh5`.
"""

from __future__ import annotations

import sys
from collections.abc import Iterator
from typing import Union
from warnings import warn

import h5py
import numpy as np
import pandas as pd

from . import lh5
from .types import (
    Array,
    ArrayOfEncodedEqualSizedArrays,  # noqa: F401
    ArrayOfEqualSizedArrays,  # noqa: F401
    FixedSizeArray,  # noqa: F401
    Histogram,  # noqa: F401
    Scalar,
    Struct,
    Table,  # noqa: F401
    VectorOfEncodedVectors,  # noqa: F401
    VectorOfVectors,
    WaveformTable,  # noqa: F401
)

LGDO = Union[Array, Scalar, Struct, VectorOfVectors]


class LH5Iterator(lh5.LH5Iterator):
    """
    .. warning::
        This class is deprecated, use :class:`lgdo.lh5.iterator.LH5Iterator`.

    """

    def __init__(
        self,
        lh5_files: str | list[str],
        groups: str | list[str],
        base_path: str = "",
        entry_list: list[int] | list[list[int]] | None = None,
        entry_mask: list[bool] | list[list[bool]] | None = None,
        field_mask: dict[str, bool] | list[str] | tuple[str] | None = None,
        buffer_len: int = 3200,
        friend: Iterator | None = None,
    ) -> None:
        warn(
            "lgdo.lh5_store has moved to a subfolder lgdo.lh5 containing LH5Iterator."
            "Please replace 'from lgdo.lh5_store import LH5Iterator' with 'from lgdo.lh5 import LH5Iterator'."
            "lgdo.lh5_store will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(
            lh5_files,
            groups,
            base_path,
            entry_list,
            entry_mask,
            field_mask,
            buffer_len,
            friend,
        )

    def write_object(
        self,
        obj: LGDO,
        name: str,
        lh5_file: str | h5py.File,
        group: str | h5py.Group = "/",
        start_row: int = 0,
        n_rows: int | None = None,
        wo_mode: str = "append",
        write_start: int = 0,
        **h5py_kwargs,
    ) -> None:
        """
        .. warning::
            This method is deprecated, use :meth:`lgdo.lh5.iterator.LH5Iterator.write`.

        """
        warn(
            "lgdo.lh5_store has moved to a subfolder lgdo.lh5 containing LH5Iterator. "
            "The object you are calling this function from uses the old LH5Iterator class."
            "Please replace 'from lgdo.lh5_store import LH5Iterator' with 'from lgdo.lh5 import LH5Iterator'."
            "lgdo.lh5_store will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.write(
            obj,
            name,
            lh5_file,
            group,
            start_row,
            n_rows,
            wo_mode,
            write_start,
            h5py_kwargs,
        )

    def read_object(
        self,
        name: str,
        lh5_file: str | h5py.File | list[str | h5py.File],
        start_row: int = 0,
        n_rows: int = sys.maxsize,
        idx: np.ndarray | list | tuple | list[np.ndarray | list | tuple] = None,
        field_mask: dict[str, bool] | list[str] | tuple[str] | None = None,
        obj_buf: LGDO = None,
        obj_buf_start: int = 0,
        decompress: bool = True,
    ) -> tuple[LGDO, int]:
        """
        .. warning::
            This method is deprecated, use :meth:`lgdo.lh5.iterator.LH5Iterator.read`.

        """
        warn(
            "lgdo.lh5_store has moved to a subfolder lgdo.lh5 containing LH5Iterator. "
            "The object you are calling this function from uses the old LH5Iterator class."
            "Please replace 'from lgdo.lh5_store import LH5Iterator' with 'from lgdo.lh5 import LH5Iterator'."
            "lgdo.lh5_store will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.read(
            name,
            lh5_file,
            start_row,
            n_rows,
            idx,
            field_mask,
            obj_buf,
            obj_buf_start,
            decompress,
        )


class LH5Store(lh5.LH5Store):
    """
    .. warning::
        This class is deprecated, use :class:`lgdo.lh5.iterator.LH5Store`.

    """

    def __init__(self, base_path: str = "", keep_open: bool = False):
        warn(
            "lgdo.lh5_store has moved to a subfolder lgdo.lh5 containing LH5Store. "
            "Please replace 'from lgdo.lh5_store import LH5Store' with 'from lgdo.lh5 import LH5Store'."
            "lgdo.lh5_store will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(base_path, keep_open)

    def read_object(
        self,
        name: str,
        lh5_file: str | h5py.File | list[str | h5py.File],
        **kwargs,
    ) -> tuple[LGDO, int]:
        """
        .. warning::
            This method is deprecated, use :meth:`lgdo.lh5.store.LH5Store.read`.

        """
        warn(
            "LH5Store.read_object() has been renamed to LH5Store.read(), "
            "Please update your code."
            "LH5Store.read_object() will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().read(self, name, lh5_file, **kwargs)

    def write_object(
        self,
        obj: LGDO,
        name: str,
        lh5_file: str | h5py.File,
        **kwargs,
    ) -> tuple[LGDO, int]:
        """
        .. warning::
            This method is deprecated, use :meth:`lgdo.lh5.store.LH5Store.write`.

        """
        warn(
            "LH5Store.write_object() has been renamed to LH5Store.write(), "
            "Please update your code."
            "LH5Store.write_object() will be removed in a future release.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().read(self, obj, name, lh5_file, **kwargs)


def load_dfs(
    f_list: str | list[str],
    par_list: list[str],
    lh5_group: str = "",
    idx_list: list[np.ndarray | list | tuple] | None = None,
) -> pd.DataFrame:
    """
    .. warning::
        This function is deprecated, use :meth:`lgdo.types.lgdo.LGDO.view_as` to
        view LGDO data as a Pandas data structure.

    """
    warn(
        "lgdo.lh5_store has moved to a subfolder lgdo.lh5. "
        "Please replace 'from lgdo.lh5_store import load_dfs' with 'from lgdo.lh5 import load_dfs'. "
        "lgdo.lh5_store will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return lh5.load_dfs(f_list, par_list, lh5_group, idx_list)


def load_nda(
    f_list: str | list[str],
    par_list: list[str],
    lh5_group: str = "",
    idx_list: list[np.ndarray | list | tuple] | None = None,
) -> dict[str, np.ndarray]:
    """
    .. warning::
        This function is deprecated, use :meth:`lgdo.types.lgdo.LGDO.view_as` to
        view LGDO data as a NumPy data structure.

    """
    warn(
        "lgdo.lh5_store has moved to a subfolder lgdo.lh5. "
        "Please replace 'from lgdo.lh5_store import load_nda' with 'from lgdo.lh5 import load_nda'. "
        "lgdo.lh5_store will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return lh5.load_nda(f_list, par_list, lh5_group, idx_list)


def ls(lh5_file: str | h5py.Group, lh5_group: str = "") -> list[str]:
    """
    .. warning::
        This function is deprecated, import :func:`lgdo.lh5.tools.ls`.

    """
    warn(
        "lgdo.lh5_store has moved to a subfolder lgdo.lh5. "
        "Please replace 'from lgdo.lh5_store import ls' with 'from lgdo.lh5 import ls'. "
        "lgdo.lh5_store will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return lh5.ls(lh5_file, lh5_group)


def show(
    lh5_file: str | h5py.Group,
    lh5_group: str = "/",
    attrs: bool = False,
    indent: str = "",
    header: bool = True,
) -> None:
    """
    .. warning::
        This function is deprecated, import :func:`lgdo.lh5.tools.show`.

    """
    warn(
        "lgdo.lh5_store has moved to a subfolder lgdo.lh5. "
        "Please replace 'from lgdo.lh5_store import show' with 'from lgdo.lh5 import show'. "
        "lgdo.lh5_store will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    lh5.show(lh5_file, lh5_group, attrs, indent, header)



"""Implements utilities for LEGEND Data Objects."""

from __future__ import annotations

import glob
import logging
import os
import string
from collections.abc import Mapping, Sequence
from typing import Any

import h5py

from .. import types
from . import _serializers
from .exceptions import LH5DecodeError

log = logging.getLogger(__name__)


def get_buffer(
    name: str,
    lh5_file: str | h5py.File | Sequence[str | h5py.File],
    size: int | None = None,
    field_mask: Mapping[str, bool] | Sequence[str] | None = None,
) -> types.LGDO:
    """Returns an LGDO appropriate for use as a pre-allocated buffer.

    Sets size to `size` if object has a size.
    """
    obj, n_rows = _serializers._h5_read_lgdo(
        lh5_file[name], n_rows=0, field_mask=field_mask
    )

    if hasattr(obj, "resize") and size is not None:
        obj.resize(new_size=size)

    return obj


def read_n_rows(name: str, h5f: str | h5py.File) -> int | None:
    """Look up the number of rows in an Array-like LGDO object on disk.

    Return ``None`` if `name` is a :class:`.Scalar` or a :class:`.Struct`.
    """
    if not isinstance(h5f, h5py.File):
        h5f = h5py.File(h5f, "r", locking=False)

    try:
        h5o = h5f[name].id
    except KeyError as e:
        msg = "not found"
        raise LH5DecodeError(msg, h5f, name) from e

    return _serializers.read.utils.read_n_rows(h5o, h5f.name, name)


def read_size_in_bytes(name: str, h5f: str | h5py.File) -> int | None:
    """Look up the size (in B) in an LGDO object in memory. Will crawl
    recursively through members of a Struct or Table
    """
    if not isinstance(h5f, h5py.File):
        h5f = h5py.File(h5f, "r", locking=False)

    try:
        h5o = h5f[name].id
    except KeyError as e:
        msg = "not found"
        raise LH5DecodeError(msg, h5f, name) from e

    return _serializers.read.utils.read_size_in_bytes(h5o, h5f.name, name)


def get_h5_group(
    group: str | h5py.Group,
    base_group: h5py.Group,
    grp_attrs: Mapping[str, Any] | None = None,
    overwrite: bool = False,
) -> h5py.Group:
    """
    Returns an existing :mod:`h5py` group from a base group or creates a
    new one. Can also set (or replace) group attributes.

    Parameters
    ----------
    group
        name of the HDF5 group.
    base_group
        HDF5 group to be used as a base.
    grp_attrs
        HDF5 group attributes.
    overwrite
        whether overwrite group attributes, ignored if `grp_attrs` is
        ``None``.
    """
    if not isinstance(group, h5py.Group):
        if group in base_group:
            group = base_group[group]
        else:
            group = base_group.create_group(group)
            if grp_attrs is not None:
                group.attrs.update(
                    {
                        k: v.encode("utf-8") if isinstance(v, str) else v
                        for k, v in grp_attrs.items()
                    }
                )
            return group
    if (
        grp_attrs is not None
        and len(set(grp_attrs.items()) ^ set(group.attrs.items())) > 0
    ):
        if not overwrite:
            msg = (
                f"Provided {grp_attrs=} are different from "
                f"existing ones {dict(group.attrs)=} but overwrite flag is not set"
            )
            raise RuntimeError(msg)

        log.debug(f"overwriting {group}.attrs...")
        for key in group.attrs:
            group.attrs.pop(key)

        group.attrs.update(
            {
                k: v.encode("utf-8") if isinstance(v, str) else v
                for k, v in grp_attrs.items()
            }
        )

    return group


def expand_vars(expr: str, substitute: dict[str, str] | None = None) -> str:
    """Expand (environment) variables.

    Note
    ----
    Malformed variable names and references to non-existing variables are left
    unchanged.

    Parameters
    ----------
    expr
        string expression, which may include (environment) variables prefixed by
        ``$``.
    substitute
        use this dictionary to substitute variables. Takes precedence over
        environment variables.
    """
    if substitute is None:
        substitute = {}

    # use provided mapping
    # then expand env variables
    return os.path.expandvars(string.Template(expr).safe_substitute(substitute))


def expand_path(
    path: str,
    substitute: dict[str, str] | None = None,
    list: bool = False,
    base_path: str | None = None,
) -> str | list:
    """Expand (environment) variables and wildcards to return absolute paths.

    Parameters
    ----------
    path
        name of path, which may include environment variables and wildcards.
    list
        if ``True``, return a list. If ``False``, return a string; if ``False``
        and a unique file is not found, raise an exception.
    substitute
        use this dictionary to substitute variables. Environment variables take
        precedence.
    base_path
        name of base path. Returned paths will be relative to base.

    Returns
    -------
    path or list of paths
        Unique absolute path, or list of all absolute paths
    """
    if base_path is not None and base_path != "":
        base_path = os.path.expanduser(os.path.expandvars(base_path))
        path = os.path.join(base_path, path)

    # first expand variables
    _path = expand_vars(path, substitute)

    # then expand wildcards
    paths = sorted(glob.glob(os.path.expanduser(_path)))

    if base_path is not None and base_path != "":
        paths = [os.path.relpath(p, base_path) for p in paths]

    if not list:
        if len(paths) == 0:
            msg = f"could not find path matching {path}"
            raise FileNotFoundError(msg)
        if len(paths) > 1:
            msg = f"found multiple paths matching {path}"
            raise FileNotFoundError(msg)

        return paths[0]

    return paths


# https://stackoverflow.com/a/1094933
def fmtbytes(num, suffix="B"):
    """Returns formatted f-string for printing human-readable number of bytes."""
    for unit in ("", "k", "M", "G", "T", "P", "E", "Z"):
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Y{suffix}"



from __future__ import annotations

import pickle
import warnings

import awkward as ak
import numpy as np
import pandas as pd
import pytest

import lgdo
from lgdo import Table


def test_init():
    tbl = Table()
    assert not tbl.size
    assert tbl.loc == 0

    tbl = Table(size=10)
    assert tbl.size == 10

    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4])),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8])),
    }

    tbl = Table(col_dict=col_dict)
    assert tbl.size == 4

    tbl = Table(size=3, col_dict=col_dict)
    assert tbl.size == 3


def test_init_nested():
    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4])),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8])),
        "c": {
            "f1": lgdo.Array([1, 2, 3, 4]),
            "f2": lgdo.Array([1, 2, 3, 4]),
        },
    }

    tbl = Table(col_dict=col_dict)
    assert isinstance(tbl.c, Table)
    assert isinstance(tbl.c.f1, lgdo.Array)
    assert tbl.c.f1 == lgdo.Array([1, 2, 3, 4])


def test_pandas_df_init():
    df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    tbl = Table(col_dict=df)
    assert sorted(tbl.keys()) == ["a", "b"]
    assert isinstance(tbl.a, lgdo.Array)
    assert isinstance(tbl.b, lgdo.Array)
    assert tbl.a == lgdo.Array([1, 2, 3, 4])
    assert tbl.b == lgdo.Array([5, 6, 7, 8])


def test_ak_array_init():
    array = ak.Array(
        {
            "a": [1, 2, 3, 4],
            "b": [[1, 2], [3], [4], [5, 6, 7]],
            "c": {"f1": [[], [5], [3, 7, 6], []], "f2": [5, 6, 7, 8]},
        }
    )
    tbl = Table(array)
    assert isinstance(tbl.a, lgdo.Array)
    assert isinstance(tbl.b, lgdo.VectorOfVectors)
    assert isinstance(tbl.c, Table)
    assert isinstance(tbl.c.f1, lgdo.VectorOfVectors)
    assert isinstance(tbl.c.f2, lgdo.Array)


def test_datatype_name():
    tbl = Table()
    assert tbl.datatype_name() == "table"


def test_push_row():
    tbl = Table()
    tbl.push_row()
    assert tbl.loc == 1


def test_is_full():
    tbl = Table(size=2)
    tbl.push_row()
    assert tbl.is_full() is False
    tbl.push_row()
    assert tbl.is_full() is True


def test_clear():
    tbl = Table()
    tbl.push_row()
    tbl.clear()
    assert tbl.loc == 0


def test_add_field():
    tbl = Table()
    tbl.add_field("a", lgdo.Array(np.array([1, 2, 3])), use_obj_size=True)
    assert tbl.size == 3

    with pytest.raises(TypeError):
        tbl.add_field("s", lgdo.Scalar(value=69))


def test_add_column():
    tbl = Table()
    tbl.add_column("a", lgdo.Array(np.array([1, 2, 3])), use_obj_size=True)
    assert tbl.size == 3
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        tbl.add_column("b", lgdo.Array(np.array([1, 2, 3, 4])))
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)


def test_join():
    tbl1 = Table(size=3)
    tbl1.add_field("a", lgdo.FixedSizeArray(np.array([1, 2, 3])))
    tbl1.add_field("b", lgdo.Array(np.array([1, 2, 3])))
    assert list(tbl1.keys()) == ["a", "b"]

    tbl2 = Table(size=3)
    tbl2.add_field("c", lgdo.Array(np.array([4, 5, 6])))
    tbl2.add_field("d", lgdo.Array(np.array([9, 9, 10])))

    tbl1.join(tbl2)
    assert list(tbl1.keys()) == ["a", "b", "c", "d"]

    tbl2.join(tbl1, cols=("a"))
    assert list(tbl2.keys()) == ["c", "d", "a"]


def test_view_as():
    tbl = Table(size=3)
    tbl.add_column("a", lgdo.Array(np.array([1, 2, 3]), attrs={"units": "m"}))
    tbl.add_column("b", lgdo.Array(np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])))
    tbl.add_column(
        "c",
        lgdo.VectorOfVectors(
            flattened_data=lgdo.Array(np.array([0, 1, 2, 3, 4, 5, 6])),
            cumulative_length=lgdo.Array(np.array([3, 4, 7])),
        ),
    )
    tbl.add_column(
        "d",
        lgdo.Table(
            col_dict={
                "a": lgdo.Array(np.array([2, 4, 6]), attrs={"units": "m"}),
                "b": lgdo.Array(np.array([[1, 1], [2, 4], [3, 9]])),
            }
        ),
    )

    df = tbl.view_as("pd", with_units=False)
    assert isinstance(df, pd.DataFrame)
    assert list(df.keys()) == ["a", "b", "c", "d_a", "d_b"]

    df = tbl.view_as("pd", with_units=True)
    assert isinstance(df, pd.DataFrame)
    assert list(df.keys()) == ["a", "b", "c", "d_a", "d_b"]
    assert df["a"].dtype == "meter"
    assert df["d_a"].dtype == "meter"

    ak_arr = tbl.view_as("ak", with_units=False)
    assert isinstance(ak_arr, ak.Array)
    assert list(ak_arr.fields) == ["a", "b", "c", "d"]

    with pytest.raises(ValueError):
        tbl.view_as("ak", with_units=True)

    with pytest.raises(TypeError):
        tbl.view_as("np")

    tbl.add_column(
        "e",
        lgdo.VectorOfVectors(
            flattened_data=lgdo.Array(np.array([0, 1, 2, 3, 4, 5, 6])),
            cumulative_length=lgdo.Array(np.array([3, 4, 7])),
            attrs={"units": "m"},
        ),
    )

    with pytest.raises(ValueError):
        tbl.view_as("pd", with_units=True)


def test_flatten():
    tbl = Table(
        col_dict={
            "a": lgdo.Array(np.array([1, 2, 3])),
            "tbl": Table(
                col_dict={
                    "b": lgdo.Array(np.array([4, 5, 6])),
                    "tbl1": Table(col_dict={"z": lgdo.Array(np.array([9, 9, 9]))}),
                }
            ),
        }
    )

    fl_tbl = tbl.flatten()
    assert sorted(fl_tbl.keys()) == ["a", "tbl__b", "tbl__tbl1__z"]


def test_remove_column():
    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4])),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8])),
        "c": lgdo.Array(nda=np.array([9, 10, 11, 12])),
    }

    tbl = Table(col_dict=col_dict)

    tbl.remove_column("a")
    assert list(tbl.keys()) == ["b", "c"]

    tbl.remove_column("c")
    assert list(tbl.keys()) == ["b"]


def test_pickle():
    col_dict = {
        "a": lgdo.Array(nda=np.array([1, 2, 3, 4])),
        "b": lgdo.Array(nda=np.array([5, 6, 7, 8])),
        "c": lgdo.Array(nda=np.array([9, 10, 11, 12])),
    }
    obj = Table(col_dict=col_dict)
    obj.attrs["attr1"] = 1

    ex = pickle.loads(pickle.dumps(obj))
    assert isinstance(ex, Table)
    assert ex.attrs["attr1"] == 1
    assert ex.attrs["datatype"] == obj.attrs["datatype"]
    for key, val in col_dict.items():
        assert ex[key] == val



r"""Data compression utilities.

This subpackage collects all LEGEND custom data compression (encoding) and
decompression (decoding) algorithms.

Available lossless waveform compression algorithms:

* :class:`.RadwareSigcompress`, a Python port of the C algorithm
  `radware-sigcompress` by D. Radford.
* :class:`.ULEB128ZigZagDiff` variable-length base-128 encoding of waveform
  differences.

All waveform compression algorithms inherit from the :class:`.WaveformCodec`
abstract class.

:func:`~.generic.encode` and :func:`~.generic.decode` provide a high-level
interface for encoding/decoding :class:`~.lgdo.LGDO`\ s.

>>> from lgdo import WaveformTable, compression
>>> wftbl = WaveformTable(...)
>>> enc_wft = compression.encode(wftable, RadwareSigcompress(codec_shift=-23768)
>>> compression.decode(enc_wft) # == wftbl
"""

from __future__ import annotations

from .base import WaveformCodec
from .generic import decode, encode
from .radware import RadwareSigcompress
from .varlen import ULEB128ZigZagDiff

__all__ = [
    "RadwareSigcompress",
    "ULEB128ZigZagDiff",
    "WaveformCodec",
    "decode",
    "encode",
]



from __future__ import annotations

import logging
import os
from inspect import signature

import h5py

from .... import compression, types
from ... import datatype, utils
from ...exceptions import LH5EncodeError
from .array import _h5_write_array
from .scalar import _h5_write_scalar
from .vector_of_vectors import _h5_write_vector_of_vectors

log = logging.getLogger(__name__)


def _h5_write_lgdo(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    assert isinstance(obj, types.LGDO)

    file_kwargs = {
        k: h5py_kwargs[k] for k in h5py_kwargs & signature(h5py.File).parameters.keys()
    }
    h5py_kwargs = {k: h5py_kwargs[k] for k in h5py_kwargs - file_kwargs.keys()}
    if wo_mode == "write_safe":
        wo_mode = "w"
    if wo_mode == "append":
        wo_mode = "a"
    if wo_mode == "overwrite":
        wo_mode = "o"
    if wo_mode == "overwrite_file":
        wo_mode = "of"
        write_start = 0
    if wo_mode == "append_column":
        wo_mode = "ac"
    if wo_mode not in ["w", "a", "o", "of", "ac"]:
        msg = f"unknown wo_mode '{wo_mode}'"
        raise LH5EncodeError(msg, lh5_file, group, name)

    # "mode" is for the h5df.File and wo_mode is for this function
    # In hdf5, 'a' is really "modify" -- in addition to appending, you can
    # change any object in the file. So we use file:append for
    # write_object:overwrite.
    if not isinstance(lh5_file, h5py.File):
        mode = "w" if wo_mode == "of" or not os.path.exists(lh5_file) else "a"
        lh5_file = h5py.File(lh5_file, mode=mode, **file_kwargs)

    log.debug(
        f"writing {obj!r}[{start_row}:{n_rows}] as "
        f"{lh5_file.filename}:{group}/{name}[{write_start}:], "
        f"mode = {wo_mode}, h5py_kwargs = {h5py_kwargs}"
    )

    group = utils.get_h5_group(group, lh5_file)

    if wo_mode == "w" and name in group:
        msg = f"can't overwrite '{name}' in wo_mode 'write_safe'"
        raise LH5EncodeError(msg, lh5_file, group, name)

    # struct, table, waveform table or histogram.
    if isinstance(obj, types.Struct):
        if (
            isinstance(obj, types.Histogram)
            and wo_mode not in ["w", "o", "of"]
            and name in group
        ):
            msg = f"can't append-write to histogram in wo_mode '{wo_mode}'"
            raise LH5EncodeError(msg, lh5_file, group, name)
        if isinstance(obj, types.Histogram) and write_start != 0:
            msg = f"can't write histogram in wo_mode '{wo_mode}' with write_start != 0"
            raise LH5EncodeError(msg, lh5_file, group, name)

        return _h5_write_struct(
            obj,
            name,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    # scalars
    if isinstance(obj, types.Scalar):
        return _h5_write_scalar(obj, name, lh5_file, group, wo_mode)

    # vector of encoded vectors
    if isinstance(
        obj, (types.VectorOfEncodedVectors, types.ArrayOfEncodedEqualSizedArrays)
    ):
        group = utils.get_h5_group(
            name, group, grp_attrs=obj.attrs, overwrite=(wo_mode == "o")
        )

        # ask not to further compress flattened_data, it is already compressed!
        obj.encoded_data.flattened_data.attrs["compression"] = None

        _h5_write_vector_of_vectors(
            obj.encoded_data,
            "encoded_data",
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

        if isinstance(obj.decoded_size, types.Scalar):
            _h5_write_scalar(
                obj.decoded_size,
                "decoded_size",
                lh5_file,
                group=group,
                wo_mode=wo_mode,
            )
        else:
            _h5_write_array(
                obj.decoded_size,
                "decoded_size",
                lh5_file,
                group=group,
                start_row=start_row,
                n_rows=n_rows,
                wo_mode=wo_mode,
                write_start=write_start,
                **h5py_kwargs,
            )

        return None

    # vector of vectors
    if isinstance(obj, types.VectorOfVectors):
        return _h5_write_vector_of_vectors(
            obj,
            name,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    # if we get this far, must be one of the Array types
    if isinstance(obj, types.Array):
        return _h5_write_array(
            obj,
            name,
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )

    msg = f"do not know how to write '{name}' of type '{type(obj).__name__}'"
    raise LH5EncodeError(msg, lh5_file, group, name)


def _h5_write_struct(
    obj,
    name,
    lh5_file,
    group="/",
    start_row=0,
    n_rows=None,
    wo_mode="append",
    write_start=0,
    **h5py_kwargs,
):
    # this works for structs and derived (tables)
    assert isinstance(obj, types.Struct)

    # In order to append a column, we need to update the
    # `struct/table{old_fields}` value in `group.attrs['datatype"]` to include
    # the new fields. One way to do this is to override `obj.attrs["datatype"]`
    # to include old and new fields. Then we can write the fields to the
    # struct/table as normal.
    if wo_mode == "ac":
        if name not in group:
            msg = "Cannot append column to non-existing struct on disk"
            raise LH5EncodeError(msg, lh5_file, group, name)

        old_group = utils.get_h5_group(name, group)
        if "datatype" not in old_group.attrs:
            msg = "Cannot append column to an existing  non-LGDO object on disk"
            raise LH5EncodeError(msg, lh5_file, group, name)

        lgdotype = datatype.datatype(old_group.attrs["datatype"])
        fields = datatype.get_struct_fields(old_group.attrs["datatype"])
        if lgdotype is not type(obj):
            msg = (
                "Trying to append columns to an object of different "
                f"type {lgdotype.__name__}!={type(obj)}"
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # If the mode is `append_column`, make sure we aren't appending
        # a table that has a column of the same name as in the existing
        # table. Also make sure that the field we are adding has the
        # same size
        if len(list(set(fields).intersection(set(obj.keys())))) != 0:
            msg = (
                f"Can't append {list(set(fields).intersection(set(obj.keys())))} "
                "column(s) to a table with the same field(s)"
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # It doesn't matter what key we access, as all fields in the old table have the same size
        if (
            isinstance(obj, types.Table)
            and old_group[next(iter(old_group.keys()))].size != obj.size
        ):
            msg = (
                f"Table sizes don't match. Trying to append column of size {obj.size} "
                f"to a table of size {old_group[next(iter(old_group.keys()))].size}."
            )
            raise LH5EncodeError(msg, lh5_file, group, name)

        # Now we can append the obj.keys() to the old fields, and then update obj.attrs.
        fields.extend(list(obj.keys()))
        obj.attrs.pop("datatype")

        obj.attrs["datatype"] = obj.datatype_name() + "{" + ",".join(fields) + "}"

        # propagating wo_mode="ac" to nested LGDOs does not make any sense
        wo_mode = "append"

        # overwrite attributes of the existing struct
        attrs_overwrite = True
    else:
        attrs_overwrite = wo_mode == "o"

    group = utils.get_h5_group(
        name,
        group,
        grp_attrs=obj.attrs,
        overwrite=attrs_overwrite,
    )
    # If the mode is overwrite, then we need to peek into the file's
    # table's existing fields. If we are writing a new table to the
    # group that does not contain an old field, we should delete that
    # old field from the file
    if wo_mode == "o":
        # Find the old keys in the group that are not present in the
        # new table's keys, then delete them
        for key in list(set(group.keys()) - set(obj.keys())):
            log.debug(f"{key} is not present in new table, deleting field")
            del group[key]

    for field in obj:
        # eventually compress waveform table values with LGDO's
        # custom codecs before writing
        # if waveformtable.values.attrs["compression"] is NOT a
        # WaveformCodec, just leave it there
        obj_fld = None
        if (
            isinstance(obj, types.WaveformTable)
            and field == "values"
            and not isinstance(obj.values, types.VectorOfEncodedVectors)
            and not isinstance(obj.values, types.ArrayOfEncodedEqualSizedArrays)
            and "compression" in obj.values.attrs
            and isinstance(obj.values.attrs["compression"], compression.WaveformCodec)
        ):
            codec = obj.values.attrs["compression"]
            obj_fld = compression.encode(obj.values, codec=codec)
        else:
            obj_fld = obj[field]

        _h5_write_lgdo(
            obj_fld,
            str(field),
            lh5_file,
            group=group,
            start_row=start_row,
            n_rows=n_rows,
            wo_mode=wo_mode,
            write_start=write_start,
            **h5py_kwargs,
        )



from __future__ import annotations

from collections.abc import Iterator
from typing import Any

import awkward as ak
import awkward_pandas as akpd
import numpy as np
import pandas as pd
from numpy.typing import NDArray

from .. import utils
from .array import Array
from .lgdo import LGDO
from .scalar import Scalar
from .vectorofvectors import VectorOfVectors


class VectorOfEncodedVectors(LGDO):
    """An array of variable-length encoded arrays.

    Used to represent an encoded :class:`.VectorOfVectors`. In addition to an
    internal :class:`.VectorOfVectors` `self.encoded_data` storing the encoded
    data, a 1D :class:`.Array` in `self.encoded_size` holds the original sizes
    of the encoded vectors.

    See Also
    --------
    .VectorOfVectors
    """

    def __init__(
        self,
        encoded_data: VectorOfVectors = None,
        decoded_size: Array = None,
        attrs: dict[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        encoded_data
            the vector of encoded vectors.
        decoded_size
            an array holding the original length of each encoded vector in
            `encoded_data`.
        attrs
            A set of user attributes to be carried along with this LGDO. Should
            include information about the codec used to encode the data.
        """
        if isinstance(encoded_data, VectorOfVectors):
            self.encoded_data = encoded_data
        elif encoded_data is None:
            self.encoded_data = VectorOfVectors(dtype="ubyte")
        else:
            msg = "encoded_data must be a valid VectorOfVectors"
            raise ValueError(msg)

        if isinstance(decoded_size, Array):
            self.decoded_size = decoded_size
        elif decoded_size is not None:
            self.decoded_size = Array(decoded_size)
        elif encoded_data is not None:
            self.decoded_size = Array(
                shape=len(encoded_data), dtype="uint32", fill_val=0
            )
        elif decoded_size is None:
            self.decoded_size = Array()

        if len(self.encoded_data) != len(self.decoded_size):
            msg = "encoded_data vs. decoded_size shape mismatch"
            raise RuntimeError(msg)

        super().__init__(attrs)

    def datatype_name(self) -> str:
        return "array"

    def form_datatype(self) -> str:
        et = utils.get_element_type(self.encoded_data)
        return "array<1>{encoded_array<1>{" + et + "}}"

    def __len__(self) -> int:
        return len(self.encoded_data)

    def __eq__(self, other: VectorOfEncodedVectors) -> bool:
        if isinstance(other, VectorOfEncodedVectors):
            return (
                self.encoded_data == other.encoded_data
                and self.decoded_size == other.decoded_size
                and self.attrs == other.attrs
            )

        return False

    def resize(self, new_size: int) -> None:
        """Resize vector along the first axis.

        See Also
        --------
        .VectorOfVectors.resize
        """
        self.encoded_data.resize(new_size)
        self.decoded_size.resize(new_size)

    def append(self, value: tuple[NDArray, int]) -> None:
        """Append a 1D encoded vector at the end.

        Parameters
        ----------
        value
            a tuple holding the encoded array and its decoded size.

        See Also
        --------
        .VectorOfVectors.append
        """
        self.encoded_data.append(value[0])
        self.decoded_size.append(value[1])

    def insert(self, i: int, value: tuple[NDArray, int]) -> None:
        """Insert an encoded vector at index `i`.

        Parameters
        ----------
        i
            the new vector will be inserted before this index.
        value
            a tuple holding the encoded array and its decoded size.

        See Also
        --------
        .VectorOfVectors.insert
        """
        self.encoded_data.insert(i, value[0])
        self.decoded_size.insert(i, value[1])

    def replace(self, i: int, value: tuple[NDArray, int]) -> None:
        """Replace the encoded vector (and decoded size) at index `i` with a new one.

        Parameters
        ----------
        i
            index of the vector to be replaced.
        value
            a tuple holding the encoded array and its decoded size.

        See Also
        --------
        .VectorOfVectors.replace
        """
        self.encoded_data.replace(i, value[0])
        self.decoded_size[i] = value[1]

    def __setitem__(self, i: int, value: tuple[NDArray, int]) -> None:
        """Set an encoded vector at index `i`.

        Parameters
        ----------
        i
            the new vector will be set at this index.
        value
            a tuple holding the encoded array and its decoded size.
        """
        self.encoded_data[i] = value[0]
        self.decoded_size[i] = value[1]

    def __getitem__(self, i: int) -> tuple[NDArray, int]:
        """Return vector at index `i`.

        Returns
        -------
        (encoded_data, decoded_size)
            the encoded array and its decoded length.
        """
        return (self.encoded_data[i], self.decoded_size[i])

    def __iter__(self) -> Iterator[tuple[NDArray, int]]:
        yield from zip(self.encoded_data, self.decoded_size)

    def __str__(self) -> str:
        string = ""
        for pos, res in enumerate(self):
            vec, size = res[0], res[1]
            if pos != 0:
                string += " "

            string += (
                np.array2string(
                    vec,
                    prefix=" ",
                    formatter={
                        "int": lambda x, vec=vec: f"0x{x:02x}"
                        if vec.dtype == np.ubyte
                        else str(x)
                    },
                )
                + f" decoded_size = {size}"
            )

            if pos < len(self.encoded_data.cumulative_length):
                string += ",\n"

        string = f"[{string}]"

        attrs = self.getattrs()
        if len(attrs) > 0:
            string += f" with attrs={attrs}"

        return string

    def __repr__(self) -> str:
        npopt = np.get_printoptions()
        np.set_printoptions(
            threshold=5,
            edgeitems=2,
            linewidth=100,
        )
        out = (
            "VectorOfEncodedVectors(encoded_data="
            + repr(self.encoded_data)
            + ", decoded_size="
            + repr(self.decoded_size)
            + ", attrs="
            + repr(self.attrs)
            + ")"
        )
        np.set_printoptions(**npopt)
        return out

    def view_as(
        self, library: str, with_units: bool = False
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        """View the encoded data as a third-party format data structure.

        This is a zero-copy or nearly zero-copy operation.

        Supported third-party formats are:

        - ``pd``: returns a :class:`pandas.DataFrame`
        - ``ak``: returns an :class:`ak.Array` (record type)

        Parameters
        ----------
        library
            format of the returned data view.
        with_units
            forward physical units to the output data.

        See Also
        --------
        .LGDO.view_as
        """
        attach_units = with_units and "units" in self.attrs

        if library == "ak":
            if attach_units:
                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            records_list = {
                "encoded_data": self.encoded_data.view_as("ak"),
                "decoded_size": np.array(self.decoded_size),
            }
            return ak.Array(records_list)

        if library == "np":
            msg = f"Format {library} is not supported for VectorOfEncodedVectors."
            raise TypeError(msg)
        if library == "pd":
            if attach_units:
                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            return pd.DataFrame(
                {
                    "encoded_data": akpd.from_awkward(self.encoded_data.view_as("ak")),
                    "decoded_size": self.decoded_size,
                }
            )

        msg = f"{library} is not a supported third-party format."
        raise ValueError(msg)


class ArrayOfEncodedEqualSizedArrays(LGDO):
    """An array of encoded arrays with equal decoded size.

    Used to represent an encoded :class:`.ArrayOfEqualSizedArrays`. In addition
    to an internal :class:`.VectorOfVectors` `self.encoded_data` storing the
    encoded data, the size of the decoded arrays is stored in a
    :class:`.Scalar` `self.encoded_size`.

    See Also
    --------
    .ArrayOfEqualSizedArrays
    """

    def __init__(
        self,
        encoded_data: VectorOfVectors = None,
        decoded_size: Scalar | int = None,
        attrs: dict[str, Any] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        encoded_data
            the vector of vectors holding the encoded data.
        decoded_size
            the length of the decoded arrays.
        attrs
            A set of user attributes to be carried along with this LGDO. Should
            include information about the codec used to encode the data.
        """
        if isinstance(encoded_data, VectorOfVectors):
            self.encoded_data = encoded_data
        elif encoded_data is None:
            self.encoded_data = VectorOfVectors(dtype="ubyte")
        else:
            msg = "encoded_data must be a valid VectorOfVectors"
            raise ValueError(msg)

        if isinstance(decoded_size, Scalar):
            self.decoded_size = decoded_size
        elif decoded_size is not None:
            self.decoded_size = Scalar(int(decoded_size))
        else:
            self.decoded_size = Scalar(0)

        super().__init__(attrs)

    def datatype_name(self) -> str:
        return "array"

    def form_datatype(self) -> str:
        et = utils.get_element_type(self.encoded_data)
        return "array_of_encoded_equalsized_arrays<1,1>{" + et + "}"

    def __len__(self) -> int:
        return len(self.encoded_data)

    def __eq__(self, other: ArrayOfEncodedEqualSizedArrays) -> bool:
        if isinstance(other, ArrayOfEncodedEqualSizedArrays):
            return (
                self.encoded_data == other.encoded_data
                and self.decoded_size == other.decoded_size
                and self.attrs == other.attrs
            )

        return False

    def resize(self, new_size: int) -> None:
        """Resize array along the first axis.

        See Also
        --------
        .VectorOfVectors.resize
        """
        self.encoded_data.resize(new_size)

    def append(self, value: NDArray) -> None:
        """Append a 1D encoded array at the end.

        See Also
        --------
        .VectorOfVectors.append
        """
        self.encoded_data.append(value)

    def insert(self, i: int, value: NDArray) -> None:
        """Insert an encoded array at index `i`.

        See Also
        --------
        .VectorOfVectors.insert
        """
        self.encoded_data.insert(i, value)

    def replace(self, i: int, value: NDArray) -> None:
        """Replace the encoded array at index `i` with a new one.

        See Also
        --------
        .VectorOfVectors.replace
        """
        self.encoded_data.replace(i, value)

    def __setitem__(self, i: int, value: NDArray) -> None:
        """Set an encoded array at index `i`."""
        self.encoded_data[i] = value

    def __getitem__(self, i: int) -> NDArray:
        """Return encoded array at index `i`."""
        return self.encoded_data[i]

    def __iter__(self) -> Iterator[NDArray]:
        yield from self.encoded_data

    def __str__(self) -> str:
        string = ""
        for pos, vec in enumerate(self):
            if pos != 0:
                string += " "

            string += np.array2string(
                vec,
                prefix=" ",
                formatter={
                    "int": lambda x, vec=vec: f"0x{x:02x}"
                    if vec.dtype == np.ubyte
                    else str(x)
                },
            )

            if pos < len(self.encoded_data.cumulative_length):
                string += ",\n"

        string = f"[{string}] decoded_size={self.decoded_size}"

        attrs = self.getattrs()
        if len(attrs) > 0:
            string += f" with attrs={attrs}"

        return string

    def __repr__(self) -> str:
        npopt = np.get_printoptions()
        np.set_printoptions(
            threshold=5,
            edgeitems=2,
            linewidth=100,
        )
        out = (
            "ArrayOfEncodedEqualSizedArrays(encoded_data="
            + repr(self.encoded_data)
            + ", decoded_size="
            + repr(self.decoded_size)
            + ", attrs="
            + repr(self.attrs)
            + ")"
        )
        np.set_printoptions(**npopt)
        return out

    def view_as(
        self, library: str, with_units: bool = False
    ) -> pd.DataFrame | np.NDArray | ak.Array:
        """View the encoded data as a third-party format data structure.

        This is nearly a zero-copy operation.

        Supported third-party formats are:

        - ``pd``: returns a :class:`pandas.DataFrame`
        - ``ak``: returns an :class:`ak.Array` (record type)

        Note
        ----
        In the view, `decoded_size` is expanded into an array.

        Parameters
        ----------
        library
            format of the returned data view.
        with_units
            forward physical units to the output data.

        See Also
        --------
        .LGDO.view_as
        """
        attach_units = with_units and "units" in self.attrs

        if library == "ak":
            if attach_units:
                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            records_list = {
                "encoded_data": self.encoded_data.view_as("ak"),
                "decoded_size": np.full(
                    len(self.encoded_data.cumulative_length), self.decoded_size.value
                ),
            }
            return ak.Array(records_list)

        if library == "np":
            msg = (
                f"Format {library} is not supported for ArrayOfEncodedEqualSizedArrays."
            )
            raise TypeError(msg)

        if library == "pd":
            if attach_units:
                msg = "Pint does not support Awkward yet, you must view the data with_units=False"
                raise ValueError(msg)

            return pd.DataFrame(
                {
                    "encoded_data": akpd.from_awkward(self.encoded_data.view_as("ak")),
                    "decoded_size": np.full(
                        len(self.encoded_data.cumulative_length),
                        self.decoded_size.value,
                    ),
                }
            )

        msg = f"{library} is not a supported third-party format."
        raise ValueError(msg)



from __future__ import annotations

import fnmatch
import glob
import logging
import os
from copy import copy
from warnings import warn

import h5py
import numpy as np
import pandas as pd
from numpy.typing import NDArray

from . import utils
from .store import LH5Store

log = logging.getLogger(__name__)


def ls(
    lh5_file: str | h5py.Group,
    lh5_group: str = "",
    recursive: bool = False,
) -> list[str]:
    """Return a list of LH5 groups in the input file and group, similar
    to ``ls`` or ``h5ls``. Supports wildcards in group names.


    Parameters
    ----------
    lh5_file
        name of file.
    lh5_group
        group to search. add a ``/`` to the end of the group name if you want to
        list all objects inside that group.
    recursive
        if ``True``, recurse into subgroups.
    """

    log.debug(
        f"Listing objects in '{lh5_file}'"
        + ("" if lh5_group == "" else f" (and group {lh5_group})")
    )

    lh5_st = LH5Store()
    # To use recursively, make lh5_file a h5group instead of a string
    if isinstance(lh5_file, str):
        lh5_file = lh5_st.gimme_file(lh5_file, "r")
        if lh5_group.startswith("/"):
            lh5_group = lh5_group[1:]

    if lh5_group == "":
        lh5_group = "*"

    # get the first group in the group path
    splitpath = lh5_group.split("/", 1)
    # filter out objects that don't match lh5_group pattern
    matchingkeys = fnmatch.filter(lh5_file.keys(), splitpath[0])

    ret = []
    # if there were no "/" in lh5_group just return the result
    if len(splitpath) == 1:
        ret = matchingkeys

    else:
        for key in matchingkeys:
            ret.extend([f"{key}/{path}" for path in ls(lh5_file[key], splitpath[1])])

    if recursive:
        rec_ret = copy(ret)
        for obj in ret:
            try:
                rec_ret += ls(lh5_file, lh5_group=f"{obj}/", recursive=True)
            except AttributeError:
                continue

        return rec_ret

    return ret


def show(
    lh5_file: str | h5py.Group,
    lh5_group: str = "/",
    attrs: bool = False,
    indent: str = "",
    header: bool = True,
    depth: int | None = None,
    detail: bool = False,
) -> None:
    """Print a tree of LH5 file contents with LGDO datatype.

    Parameters
    ----------
    lh5_file
        the LH5 file.
    lh5_group
        print only contents of this HDF5 group.
    attrs
        print the HDF5 attributes too.
    indent
        indent the diagram with this string.
    header
        print `lh5_group` at the top of the diagram.
    depth
        maximum tree depth of groups to print
    detail
        whether to print additional information about how the data is stored

    Examples
    --------
    >>> from lgdo import show
    >>> show("file.lh5", "/geds/raw")
    /geds/raw
    ├── channel · array<1>{real}
    ├── energy · array<1>{real}
    ├── timestamp · array<1>{real}
    ├── waveform · table{t0,dt,values}
    │   ├── dt · array<1>{real}
    │   ├── t0 · array<1>{real}
    │   └── values · array_of_equalsized_arrays<1,1>{real}
    └── wf_std · array<1>{real}
    """
    # check tree depth if we are using it
    if depth is not None and depth <= 0:
        return

    # open file
    if isinstance(lh5_file, str):
        lh5_file = h5py.File(utils.expand_path(lh5_file), "r", locking=False)

    # go to group
    if lh5_group != "/":
        lh5_file = lh5_file[lh5_group]

    if header:
        print(f"\033[1m{lh5_group}\033[0m")  # noqa: T201

    # get an iterator over the keys in the group
    it = iter(lh5_file)
    key = None

    # make sure there is actually something in this file/group
    try:
        key = next(it)  # get first key
    except StopIteration:
        print(f"{indent}└──  empty")  # noqa: T201
        return

    # loop over keys
    while True:
        val = lh5_file[key]
        # we want to print the LGDO datatype
        dtype = val.attrs.get("datatype", default="no datatype")
        if dtype == "no datatype" and isinstance(val, h5py.Group):
            dtype = "HDF5 group"

        _attrs = ""
        if attrs:
            attrs_d = dict(val.attrs)
            attrs_d.pop("datatype", "")
            _attrs = "── " + str(attrs_d) if attrs_d else ""

        # is this the last key?
        killme = False
        try:
            k_new = next(it)  # get next key
        except StopIteration:
            char = "└──"
            killme = True  # we'll have to kill this loop later
        else:
            char = "├──"

        print(f"{indent}{char} \033[1m{key}\033[0m · {dtype} {_attrs}")  # noqa: T201

        if detail and isinstance(val, h5py.Dataset):
            char = "|       "
            if killme:
                char = "        "
            toprint = f"{indent}{char}"
            try:
                toprint += f"\033[3mdtype\033[0m={val.dtype}"
                toprint += f", \033[3mshape\033[0m={val.shape}"
                toprint += f", \033[3mnbytes\033[0m={utils.fmtbytes(val.nbytes)}"
                if (chunkshape := val.chunks) is None:
                    toprint += ", \033[3mnumchunks\033[0m=contiguous"
                else:
                    toprint += f", \033[3mnumchunks\033[0m={val.id.get_num_chunks()}"
                    toprint += f", \033[3mchunkshape\033[0m={chunkshape}"
                toprint += ", \033[3mfilters\033[0m="

                numfilters = val.id.get_create_plist().get_nfilters()
                if numfilters == 0:
                    toprint += "None"
                else:
                    toprint += "("
                    for i in range(numfilters):
                        thisfilter = val.id.get_create_plist().get_filter(i)[3].decode()
                        if "lz4" in thisfilter:
                            thisfilter = "lz4"
                        toprint += f"{thisfilter},"
                    toprint += ")"

            except TypeError:
                toprint += "(scalar)"

            print(toprint)  # noqa: T201

        # if it's a group, call this function recursively
        if isinstance(val, h5py.Group):
            show(
                val,
                indent=indent + ("    " if killme else "│   "),
                header=False,
                attrs=attrs,
                depth=depth - 1 if depth else None,
                detail=detail,
            )

        # break or move to next key
        if killme:
            break

        key = k_new


def load_nda(
    f_list: str | list[str],
    par_list: list[str],
    lh5_group: str = "",
    idx_list: list[NDArray | list | tuple] | None = None,
) -> dict[str, NDArray]:
    r"""Build a dictionary of :class:`numpy.ndarray`\ s from LH5 data.

    Given a list of files, a list of LH5 table parameters, and an optional
    group path, return a NumPy array with all values for each parameter.

    Parameters
    ----------
    f_list
        A list of files. Can contain wildcards.
    par_list
        A list of parameters to read from each file.
    lh5_group
        group path within which to find the specified parameters.
    idx_list
        for fancy-indexed reads. Must be one index array for each file in
        `f_list`.

    Returns
    -------
    par_data
        A dictionary of the parameter data keyed by the elements of `par_list`.
        Each entry contains the data for the specified parameter concatenated
        over all files in `f_list`.
    """
    warn(
        "load_nda() is deprecated. "
        "Please replace it with LH5Store.read(...).view_as('np'), "
        "or just read_as(..., 'np'). "
        "load_nda() will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )

    if isinstance(f_list, str):
        f_list = [f_list]
        if idx_list is not None:
            idx_list = [idx_list]
    if idx_list is not None and len(f_list) != len(idx_list):
        msg = f"f_list length ({len(f_list)}) != idx_list length ({len(idx_list)})!"
        raise ValueError(msg)

    # Expand wildcards
    f_list = [f for f_wc in f_list for f in sorted(glob.glob(os.path.expandvars(f_wc)))]

    sto = LH5Store()
    par_data = {par: [] for par in par_list}
    for ii, ff in enumerate(f_list):
        f = sto.gimme_file(ff, "r")
        for par in par_list:
            if f"{lh5_group}/{par}" not in f:
                msg = f"'{lh5_group}/{par}' not in file {ff}"
                raise RuntimeError(msg)

            if idx_list is None:
                data, _ = sto.read(f"{lh5_group}/{par}", f)
            else:
                data, _ = sto.read(f"{lh5_group}/{par}", f, idx=idx_list[ii])
            if not data:
                continue
            par_data[par].append(data.nda)
    return {par: np.concatenate(par_data[par]) for par in par_list}


def load_dfs(
    f_list: str | list[str],
    par_list: list[str],
    lh5_group: str = "",
    idx_list: list[NDArray | list | tuple] | None = None,
) -> pd.DataFrame:
    """Build a :class:`pandas.DataFrame` from LH5 data.

    Given a list of files (can use wildcards), a list of LH5 columns, and
    optionally the group path, return a :class:`pandas.DataFrame` with all
    values for each parameter.

    See Also
    --------
    :func:`load_nda`

    Returns
    -------
    dataframe
        contains columns for each parameter in `par_list`, and rows containing
        all data for the associated parameters concatenated over all files in
        `f_list`.
    """
    warn(
        "load_dfs() is deprecated. "
        "Please replace it with LH5Store.read(...).view_as('pd'), "
        "or just read_as(..., 'pd'). "
        "load_dfs() will be removed in a future release.",
        DeprecationWarning,
        stacklevel=2,
    )
    return pd.DataFrame(
        load_nda(f_list, par_list, lh5_group=lh5_group, idx_list=idx_list)
    )



from __future__ import annotations

import logging
import typing
from warnings import warn

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from ..types import Array, Scalar, Struct, VectorOfVectors
from ..units import default_units_registry as ureg
from .store import LH5Store
from .utils import expand_path

LGDO = typing.Union[Array, Scalar, Struct, VectorOfVectors]


class LH5Iterator(typing.Iterator):
    """
    A class for iterating through one or more LH5 files, one block of entries
    at a time. This also accepts an entry list/mask to enable event selection,
    and a field mask.

    This can be used as an iterator:

    >>> for lh5_obj, i_entry, n_rows in LH5Iterator(...):
    >>>    # do the thing!

    This is intended for if you are reading a large quantity of data. This
    will ensure that you traverse files efficiently to minimize caching time
    and will limit your memory usage (particularly when reading in waveforms!).
    The ``lh5_obj`` that is read by this class is reused in order to avoid
    reallocation of memory; this means that if you want to hold on to data
    between reads, you will have to copy it somewhere!

    When defining an LH5Iterator, you must give it a list of files and the
    hdf5 groups containing the data tables you are reading. You may also
    provide a field mask, and an entry list or mask, specifying which entries
    to read from the files. You may also pair it with a friend iterator, which
    contains a parallel group of files which will be simultaneously read.
    In addition to accessing requested data via ``lh5_obj``, several
    properties exist to tell you where that data came from:

    - lh5_it.current_local_entries: get the entry numbers relative to the
      file the data came from
    - lh5_it.current_global_entries: get the entry number relative to the
      full dataset
    - lh5_it.current_files: get the file name corresponding to each entry
    - lh5_it.current_groups: get the group name corresponding to each entry

    This class can also be used either for random access:

    >>> lh5_obj, n_rows = lh5_it.read(i_entry)

    to read the block of entries starting at i_entry. In case of multiple files
    or the use of an event selection, i_entry refers to a global event index
    across files and does not count events that are excluded by the selection.
    """

    def __init__(
        self,
        lh5_files: str | list[str],
        groups: str | list[str] | list[list[str]],
        base_path: str = "",
        entry_list: list[int] | list[list[int]] | None = None,
        entry_mask: list[bool] | list[list[bool]] | None = None,
        field_mask: dict[str, bool] | list[str] | tuple[str] | None = None,
        buffer_len: int = "100*MB",
        file_cache: int = 10,
        file_map: NDArray[int] = None,
        friend: typing.Iterator | None = None,
    ) -> None:
        """
        Parameters
        ----------
        lh5_files
            file or files to read from. May include wildcards and environment
            variables.
        groups
            HDF5 group(s) to read. If a list of strings is provided, use
            same groups for each file. If a list of lists is provided, size
            of outer list must match size of file list, and each inner list
            will apply to a single file (or set of wildcarded files)
        entry_list
            list of entry numbers to read. If a nested list is provided,
            expect one top-level list for each file, containing a list of
            local entries. If a list of ints is provided, use global entries.
        entry_mask
            mask of entries to read. If a list of arrays is provided, expect
            one for each file. Ignore if a selection list is provided.
        field_mask
            mask of which fields to read. See :meth:`LH5Store.read` for
            more details.
        buffer_len
            number of entries to read at a time while iterating through files.
        file_cache
            maximum number of files to keep open at a time
        file_map
            cumulative file/group entries. This can be provided on construction
            to speed up random or sparse access; otherwise, we sequentially
            read the size of each group. WARNING: no checks for accuracy are
            performed so only use this if you know what you are doing!
        friend
            a \"friend\" LH5Iterator that will be read in parallel with this.
            The friend should have the same length and entry list. A single
            LH5 table containing columns from both iterators will be returned.
            Note that buffer_len will be set to the minimum of the two.
        """
        self.lh5_st = LH5Store(base_path=base_path, keep_open=file_cache)

        # List of files, with wildcards and env vars expanded
        if isinstance(lh5_files, str):
            lh5_files = [lh5_files]
        elif not isinstance(lh5_files, (list, set, tuple)):
            msg = "lh5_files must be a string or list of strings"
            raise ValueError(msg)

        if isinstance(groups, str):
            groups = [[groups]] * len(lh5_files)
        elif not isinstance(groups, list):
            msg = "group must be a string or appropriate list"
            raise ValueError(msg)
        elif all(isinstance(g, str) for g in groups):
            groups = [groups] * len(lh5_files)
        elif len(groups) == len(lh5_files) and all(
            isinstance(gr_list, (list, set, tuple)) for gr_list in groups
        ):
            pass
        else:
            msg = "group must be a string or appropriate list"
            raise ValueError(msg)

        if len(groups) != len(lh5_files):
            msg = "lh5_files and groups must have same length"
            raise ValueError(msg)

        # make flattened outer-product-like list of files and groups
        self.lh5_files = []
        self.groups = []
        for f, g in zip(lh5_files, groups):
            for f_exp in expand_path(f, list=True, base_path=base_path):
                self.lh5_files += [f_exp] * len(g)
                self.groups += list(g)

        if entry_list is not None and entry_mask is not None:
            msg = "entry_list and entry_mask arguments are mutually exclusive"
            raise ValueError(msg)

        # Map to last row in each file
        if file_map is None:
            self.file_map = np.full(len(self.lh5_files), np.iinfo("q").max, "q")
        else:
            self.file_map = np.array(file_map)

        # Map to last iterator entry for each file
        self.entry_map = np.full(len(self.lh5_files), np.iinfo("q").max, "q")
        self.buffer_len = buffer_len

        if len(self.lh5_files) > 0:
            f = self.lh5_files[0]
            g = self.groups[0]
            n_rows = self.lh5_st.read_n_rows(g, f)

            if isinstance(self.buffer_len, str):
                self.buffer_len = ureg.Quantity(buffer_len)
            if isinstance(self.buffer_len, ureg.Quantity):
                self.buffer_len = int(
                    self.buffer_len
                    / (self.lh5_st.read_size_in_bytes(g, f) * ureg.B)
                    * n_rows
                )

            self.lh5_buffer = self.lh5_st.get_buffer(
                g,
                f,
                size=self.buffer_len,
                field_mask=field_mask,
            )
            if file_map is None:
                self.file_map[0] = n_rows
        else:
            msg = f"can't open any files from {lh5_files}"
            raise RuntimeError(msg)

        self.n_rows = 0
        self.current_i_entry = 0
        self.next_i_entry = 0

        self.field_mask = field_mask

        # List of entry indices from each file
        self.local_entry_list = None
        self.global_entry_list = None
        if entry_list is not None:
            entry_list = list(entry_list)
            if isinstance(entry_list[0], int):
                self.local_entry_list = [None] * len(self.file_map)
                self.global_entry_list = np.array(entry_list, "q")
                self.global_entry_list.sort()

            else:
                self.local_entry_list = [[]] * len(self.file_map)
                for i_file, local_list in enumerate(entry_list):
                    self.local_entry_list[i_file] = np.array(local_list, "q")
                    self.local_entry_list[i_file].sort()

        elif entry_mask is not None:
            # Convert entry mask into an entry list
            if isinstance(entry_mask, pd.Series):
                entry_mask = entry_mask.to_numpy()
            if isinstance(entry_mask, np.ndarray):
                self.local_entry_list = [None] * len(self.file_map)
                self.global_entry_list = np.nonzero(entry_mask)[0]
            else:
                self.local_entry_list = [[]] * len(self.file_map)
                for i_file, local_mask in enumerate(entry_mask):
                    self.local_entry_list[i_file] = np.nonzero(local_mask)[0]

        # Attach the friend
        if friend is not None:
            if not isinstance(friend, typing.Iterator):
                msg = "Friend must be an Iterator"
                raise ValueError(msg)

            # set buffer_lens to be equal
            if self.buffer_len < friend.buffer_len:
                friend.buffer_len = self.buffer_len
                friend.lh5_buffer.resize(self.buffer_len)
            elif self.buffer_len > friend.buffer_len:
                self.buffer_len = friend.buffer_len
                self.lh5_buffer.resize(friend.buffer_len)

            self.lh5_buffer.join(friend.lh5_buffer)
        self.friend = friend

    def _get_file_cumlen(self, i_file: int) -> int:
        """Helper to get cumulative file length of file"""
        if i_file < 0:
            return 0
        fcl = self.file_map[i_file]

        # if we haven't already calculated, calculate for all files up to i_file
        if fcl == np.iinfo("q").max:
            i_start = np.searchsorted(self.file_map, np.iinfo("q").max)
            fcl = self.file_map[i_start - 1] if i_start > 0 else 0

            for i in range(i_start, i_file + 1):
                fcl += self.lh5_st.read_n_rows(self.groups[i], self.lh5_files[i])
                self.file_map[i] = fcl
        return fcl

    @property
    def current_entry(self) -> int:
        "deprecated alias for current_i_entry"
        warn(
            "current_entry has been renamed to current_i_entry.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.current_i_entry

    def _get_file_cumentries(self, i_file: int) -> int:
        """Helper to get cumulative iterator entries in file"""
        if i_file < 0:
            return 0
        n = self.entry_map[i_file]

        # if we haven't already calculated, calculate for all files up to i_file
        if n == np.iinfo("q").max:
            i_start = np.searchsorted(self.entry_map, np.iinfo("q").max)
            n = self.entry_map[i_start - 1] if i_start > 0 else 0

            for i in range(i_start, i_file + 1):
                elist = self.get_file_entrylist(i)
                fcl = self._get_file_cumlen(i)
                if elist is None:
                    # no entry list provided
                    n = fcl
                else:
                    n += len(elist)
                    # check that file entries fall inside of file
                    if len(elist) > 0 and elist[-1] >= fcl:
                        logging.warning(f"Found entries out of range for file {i}")
                        n += np.searchsorted(elist, fcl, "right") - len(elist)
                self.entry_map[i] = n
        return n

    def get_file_entrylist(self, i_file: int) -> np.ndarray:
        """Helper to get entry list for file"""
        # If no entry list is provided
        if self.local_entry_list is None:
            return None

        elist = self.local_entry_list[i_file]
        if elist is None:
            # Get local entrylist for this file from global entry list
            f_start = self._get_file_cumlen(i_file - 1)
            f_end = self._get_file_cumlen(i_file)
            i_start = self._get_file_cumentries(i_file - 1)
            i_stop = np.searchsorted(self.global_entry_list, f_end, "right")
            elist = np.array(self.global_entry_list[i_start:i_stop], "q") - f_start
            self.local_entry_list[i_file] = elist
        return elist

    def get_global_entrylist(self) -> np.ndarray:
        """Get global entry list, constructing it if needed"""
        if self.global_entry_list is None and self.local_entry_list is not None:
            self.global_entry_list = np.zeros(len(self), "q")
            for i_file in range(len(self.lh5_files)):
                i_start = self._get_file_cumentries(i_file - 1)
                i_stop = self._get_file_cumentries(i_file)
                f_start = self._get_file_cumlen(i_file - 1)
                self.global_entry_list[i_start:i_stop] = (
                    self.get_file_entrylist(i_file) + f_start
                )
        return self.global_entry_list

    def read(self, i_entry: int) -> tuple[LGDO, int]:
        """Read the nextlocal chunk of events, starting at i_entry. Return the
        LH5 buffer and number of rows read."""
        self.n_rows = 0
        i_file = np.searchsorted(self.entry_map, i_entry, "right")

        # if file hasn't been opened yet, search through files
        # sequentially until we find the right one
        if i_file < len(self.lh5_files) and self.entry_map[i_file] == np.iinfo("q").max:
            while i_file < len(self.lh5_files) and i_entry >= self._get_file_cumentries(
                i_file
            ):
                i_file += 1

        if i_file == len(self.lh5_files):
            return (self.lh5_buffer, self.n_rows)
        local_i_entry = i_entry - self._get_file_cumentries(i_file - 1)

        while self.n_rows < self.buffer_len and i_file < len(self.file_map):
            # Loop through files
            local_idx = self.get_file_entrylist(i_file)
            if local_idx is not None and len(local_idx) == 0:
                i_file += 1
                local_i_entry = 0
                continue

            i_local = local_i_entry if local_idx is None else local_idx[local_i_entry]
            self.lh5_buffer, n_rows = self.lh5_st.read(
                self.groups[i_file],
                self.lh5_files[i_file],
                start_row=i_local,
                n_rows=self.buffer_len - self.n_rows,
                idx=local_idx,
                field_mask=self.field_mask,
                obj_buf=self.lh5_buffer,
                obj_buf_start=self.n_rows,
            )

            self.n_rows += n_rows
            i_file += 1
            local_i_entry = 0

        self.current_i_entry = i_entry

        if self.friend is not None:
            self.friend.read(i_entry)

        return (self.lh5_buffer, self.n_rows)

    def reset_field_mask(self, mask):
        """Replaces the field mask of this iterator and any friends with mask"""
        self.field_mask = mask
        if self.friend is not None:
            self.friend.reset_field_mask(mask)

    @property
    def current_local_entries(self) -> NDArray[int]:
        """Return list of local file entries in buffer"""
        cur_entries = np.zeros(self.n_rows, dtype="int32")
        i_file = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        file_start = self._get_file_cumentries(i_file - 1)
        i_local = self.current_i_entry - file_start
        i = 0

        while i < len(cur_entries):
            # number of entries to read from this file
            file_end = self._get_file_cumentries(i_file)
            n = min(file_end - file_start - i_local, len(cur_entries) - i)
            entries = self.get_file_entrylist(i_file)

            if entries is None:
                cur_entries[i : i + n] = np.arange(i_local, i_local + n)
            else:
                cur_entries[i : i + n] = entries[i_local : i_local + n]

            i_file += 1
            file_start = file_end
            i_local = 0
            i += n

        return cur_entries

    @property
    def current_global_entries(self) -> NDArray[int]:
        """Return list of local file entries in buffer"""
        cur_entries = np.zeros(self.n_rows, dtype="int32")
        i_file = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        file_start = self._get_file_cumentries(i_file - 1)
        i_local = self.current_i_entry - file_start
        i = 0

        while i < len(cur_entries):
            # number of entries to read from this file
            file_end = self._get_file_cumentries(i_file)
            n = min(file_end - file_start - i_local, len(cur_entries) - i)
            entries = self.get_file_entrylist(i_file)

            if entries is None:
                cur_entries[i : i + n] = self._get_file_cumlen(i_file - 1) + np.arange(
                    i_local, i_local + n
                )
            else:
                cur_entries[i : i + n] = (
                    self._get_file_cumlen(i_file - 1) + entries[i_local : i_local + n]
                )

            i_file += 1
            file_start = file_end
            i_local = 0
            i += n

        return cur_entries

    @property
    def current_files(self) -> NDArray[str]:
        """Return list of file names for entries in buffer"""
        cur_files = np.zeros(self.n_rows, dtype=object)
        i_file = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        file_start = self._get_file_cumentries(i_file - 1)
        i_local = self.current_i_entry - file_start
        i = 0

        while i < len(cur_files):
            # number of entries to read from this file
            file_end = self._get_file_cumentries(i_file)
            n = min(file_end - file_start - i_local, len(cur_files) - i)
            cur_files[i : i + n] = self.lh5_files[i_file]

            i_file += 1
            file_start = file_end
            i_local = 0
            i += n

        return cur_files

    @property
    def current_groups(self) -> NDArray[str]:
        """Return list of group names for entries in buffer"""
        cur_groups = np.zeros(self.n_rows, dtype=object)
        i_file = np.searchsorted(self.entry_map, self.current_i_entry, "right")
        file_start = self._get_file_cumentries(i_file - 1)
        i_local = self.current_i_entry - file_start
        i = 0

        while i < len(cur_groups):
            # number of entries to read from this file
            file_end = self._get_file_cumentries(i_file)
            n = min(file_end - file_start - i_local, len(cur_groups) - i)
            cur_groups[i : i + n] = self.groups[i_file]

            i_file += 1
            file_start = file_end
            i_local = 0
            i += n

        return cur_groups

    def __len__(self) -> int:
        """Return the total number of entries."""
        return (
            self._get_file_cumentries(len(self.lh5_files) - 1)
            if len(self.entry_map) > 0
            else 0
        )

    def __iter__(self) -> typing.Iterator:
        """Loop through entries in blocks of size buffer_len."""
        self.current_i_entry = 0
        self.next_i_entry = 0
        return self

    def __next__(self) -> tuple[LGDO, int, int]:
        """Read next buffer_len entries and return lh5_table, iterator entry
        and n_rows read."""
        buf, n_rows = self.read(self.next_i_entry)
        self.next_i_entry = self.current_i_entry + n_rows
        if n_rows == 0:
            raise StopIteration
        return (buf, self.current_i_entry, n_rows)
